:- module(_,_,[assertions,regtypes,fsyntax]).

:- use_module( library( autoconfig ) ).
:- use_module( library( terms      ) , [ atom_concat/2 ] ).
:- use_module( library( system     ) , [ current_env/2
	                               , get_pwnam/1
	                               , get_grnam/1 ] ).


%% Should add extrafiles, etc.

% ----------------------------------------------------------------------------

:- comment(filetype,user).

% ----------------------------------------------------------------------------

:- comment(title,"Configuration File for Document Generation").

:- comment(module,"This is a sample configuration file for @apl{lpdoc}
   (as well as the documentation on configuration).  The defaults
   listed are typically suggestions and/or the ones used for local
   installation in the CLIP group machines.  These settings should be
   changed to suit your application. The set of variables are grouped
   by ").

:- comment(author,"Manuel Hermenegildo, David Trallero Mena").

% ----------------------------------------------------------------------------
% DIRECTORIES
% ----------------------------------------------------------------------------

:- pred package_root_dir( PackageRootDir ) => atm( PackageRootDir )

# "@var{PackageRootDir} defines the root directory from where the
  package actions (generations of tars, documentation, version ...)
  will take place (do not forget the / at the end of the path).

  Example: package_root_dir := '/home/clip/Systems/CiaoDE/'.".

% REVIEW
package_root_dir := '/home/dtm/Systems/WebDB2.0/'.

% ----------------------------------------------------------------------------
% NAMES
% ----------------------------------------------------------------------------


% PACKAGE NAME
% ============

:- pred package_name( Name ) => atm(Name)

# "@var{Name} is the name of the package. 

  Example: package_name := 'CiaoDE'.".

% REVIEW
package_name := 'WebDB2.0'.



% PACKAGE VERSION
% ===============

:- pred package_version( Version ) => atm( Version )

# "@var{Version} is the package version.

  Example: package_version := '1.15.3-r(5523)'.".

package_version( Version ) :-
	package_root_dir( PRD ),
	get_version_string(  PRD , V ),
	get_patch_string(    PRD , P ),
	get_revision_string( PRD , R ),
	atom_codes( VA , V ),
	atom_codes( PA , P ),
	atom_codes( RA , R ),
	atom_concat( [VA,'.', PA, '#', RA] , Version ).


% PACKAGE VERSION NICE
% ====================

:- pred package_version_nice( NiceVersion ) => atm( NiceVersion )

# "@var{Version} defines the package version that would be presented
  to the user in a distribution website or manual installation. It is
  the aesthetic version of @pref{package_version/1}.

  Example: package_version := '1.15.3'".

package_version_nice( NiceVersion ) :-
	package_root_dir( PRD ),
	get_version_string(  PRD , V ),
	get_patch_string(    PRD , P ),
	atom_codes( VA , V ),
	atom_codes( PA , P ),
	atom_concat( [VA,'.', PA] , NiceVersion ).



% PACKAGE NAME VERSION
% ====================

:- pred package_name_version( NameVersion ) => atm( NameVersion )

# "@var{NameVersion} is the concatenation of the package name and its
  version. It is used as a title in a distribution website or manual
  installation.

  Example: package_name_version( 'CiaoDE-1.15.2' ).".

package_name_version( NameVersion ) :-
	package_name( Name ),
	package_version_nice( Version ),
	atom_concat( [Name , '-' , Version] , NameVersion ).




% ----------------------------------------------------------------------------
% Here comes variables that defines how to build the documentation.
% These variables include: paths, mainfile, file options, doc formats, 
% buildroot, permissions and install paths
% ----------------------------------------------------------------------------


% ----------------------------------------------------------------------------
% PATHS
% ----------------------------------------------------------------------------

:- pred filepath/1 => dirpath

# "Defines the directories where the @tt{.pl} files to be documented
   can be found.  You also need to specify all the paths containing
   files which are used by the files being documented. For example,
   the paths to files included by an @tt{@@include} command or to
   figures.

   Example: filepath := '/home/clip/Systems/ciao/doc' |
                        '/home/clip/Systems/ciao/doc/common'.".
%   @includedef{filepath/1}".

% REVIEW
filepath := '/home/dtm/Systems/WebDB2.0/doc' |
	    '/home/dtm/Systems/WebDB2.0/src' |
            '/home/dtm/Systems/WebDB2.0/doc/fig'.

:- comment(dirpath/1, "An atom describing a path to a
   directory. Should be a full, explicit path (i.e., not containing
   environment variables).").

:- regtype dirpath(P) # "@var{P} is a full path to a directory.".

dirpath(P) :- atm(P).


:- regtype filename(P) # "@var{P} is a full path to a file.".

filename(P) :- atm(P).


:- pred systempath/1 => dirpath

# "Defines the @bf{system} directories where @tt{.pl} files used are.
   You also need to specify all the paths containing files which are
   used by the files being documented.  For example, library
   subdirectories containing files used in the documentation should
   also be included, as well as figures, @tt{@@include}s, etc.  You
   can put these in @pred{filepath/1} instead. Putting them here only
   means that the corresponding files will be assumed to be
   @em{system} files (and labelled as such) in the documentation.

   Example: @includedef{systempath/1}".

% REVIEW
systempath := '/home/clip/Systems/CiaoDE/trunk/ciao/lib/'     |
              '/home/clip/Systems/CiaoDE/trunk/ciao/library/' |
              % For Copyright.Manuals
              '/home/clip/Systems/CiaoDE/trunk/ciao/doc/common/'.

:- pred pathsfile(File) # "The argument denotes a file containing path
   alias definitions and oher compiler-related info used when
   compiling the files to be documented. This has the same
   functionality as the @tt{-u} option in @apl{ciaoc}. Simply leave
   undefined if you do not use path aliases, etc. Leave it commented
   for no path alias definition.

   Example: pathsfile := 'mypaths.pl' ". % @includedef{pathsfile/1}".

%% pathsfile := 'mypaths.pl'.

% ----------------------------------------------------------------------------
% The main file
% ----------------------------------------------------------------------------

:- pred mainfile(Main) => sourcename

# "Defines the file which is to be the main file of the manual, i.e.,
   the file which determines the manual's cover page, and first
   chappter.  It should not include a path unless the path should
   appear in the documentation.

   Example: @includedef{mainfile/1}".

% REVIEW
mainfile := 'WebDB'.

% ----------------------------------------------------------------------------
% The component files
% ----------------------------------------------------------------------------

:- pred component(Comp) => sourcename

# "Defines the files which are to be the used as components, i.e.,
   which will constitute the subsequent chapters of the manual. These
   can be @tt{.pl} files or @tt{.src} files (manually produced files
   in @apl{texinfo} format).  Include a path only if it should appear
   in the documentation, i.e., for sublibraries. Leave it commented if
   your application has no components.

   Example: component := 'example_module'|'comments'". 
% @includedef{component/1}".

% component := 'example_module'|'comments'.

% ----------------------------------------------------------------------------
% Setting processing options for the different files
% ----------------------------------------------------------------------------

:- pred fileoption(File,Option) :: sourcename * supported_option

# "@var{Option} is a processing option which should be activated when
   processing file @var{File}. Type @tt{lpdoc -h} to get the list of
   options.

   Example: @includedef{fileoption/2}
   @noindent sets the options above for all files.

   In general, selecting none of these options generates the most
   verbose manuals, i.e., each option generally supresses the
   production of a certain kind of output (on the other hand,
   @tt{'-v'} selects verbose output from ciaoc when processing the
   file). ".

fileoption(_) := '-nobugs'|'-noengmods'. 


:- regtype supported_option/1 # 
	"Possible options: @includedef{supported_option/1}".

supported_option := '-v'.
supported_option := '-nobugs'.
supported_option := '-noauthors'.
supported_option := '-noversion'.
supported_option := '-nochangelog'.
supported_option := '-nopatches'.
supported_option := '-modes'.
supported_option := '-headprops'.
supported_option := '-literalprops'.
supported_option := '-nopropnames'.
supported_option := '-noundefined'.
supported_option := '-nopropsepln'.
supported_option := '-norefs'.
supported_option := '-nobullet'.
supported_option := '-nosysmods'.
supported_option := '-noengmods'.
supported_option := '-noisoline'.
supported_option := '-propmods'.
supported_option := '-shorttoc'.

% ----------------------------------------------------------------------------
% Default document formats
% ----------------------------------------------------------------------------

:- pred docformat(Format) => supported_format

# "Defines the documentation formats to be generated by default when
   running @apl{lpdoc}, among the following (@em{they should be kept
   in this order}): @includedef{docformat/1}".

docformat := texi.       
docformat := ps.
docformat := pdf.
docformat := manl.
docformat := info.
docformat := html.

:- regtype supported_format/1 
# "Available formats: @includedef{supported_format/1}".

supported_format := texi.       
supported_format := dvi.
supported_format := ps.
supported_format := pdf.
supported_format := ascii.
supported_format := manl.
supported_format := info.
supported_format := html.

% ----------------------------------------------------------------------------
% Indices to be generated
% ----------------------------------------------------------------------------

:- pred index(Format) => supported_index

# "Defines the indices to be generated by default when running
   @apl{lpdoc}, among the following: 

   @noindent @tt{concept lib apl pred prop regtype decl op modedef file global}

   Selecting @tt{all} generates all the supported indices However,
   note that this (as well as selecting many indices explicitely)
   exceeds the capacity of most texinfo installations. A good set of
   indices is:

   @includedef{index/1}".

index := concept.
index := pred.
index := prop.
index := regtype.
index := modedef.
index := global.

:- regtype supported_index/1 
# "Supported indexes: @includedef{supported_index/1}".

supported_index := concept.
supported_index := lib.
supported_index := apl.
supported_index := pred.
supported_index := prop.
supported_index := regtype.
supported_index := decl.
supported_index := op.
supported_index := modedef.
supported_index := file.
supported_index := global.
supported_index := all.

% ----------------------------------------------------------------------------
% References
% ----------------------------------------------------------------------------

:- pred bibfile(Format) => filename

# "If you are using bibliographic references, define in this way the
   @tt{.bib} files to be used to find them. Defines the indices to be
   generated by default when running @apl{lpdoc}, among the following:

   Example: @includedef{bibfile/1}".

bibfile := '/home/clip/bibtex/clip/clip'.
bibfile := '/home/clip/bibtex/clip/others'.


% ----------------------------------------------------------------------------
% Other settings
% ----------------------------------------------------------------------------

:- pred startpage(PageNumber) => int

# "Setting this to a different value allows changing the page number of
   the first page of the manual. This can be useful if the manual is to
   be included in a larger document or set of manuals.
   Typically, this should be an odd number.

   Example: @includedef{startpage/1}".

startpage := 1.

:- pred papertype(PageNumber) => supported_papertype

# "Selects the type of paper/format for printed documents.  See also
   the @tt{-onesided} and @tt{-twosided} options for the main file.

   Example: @includedef{papertype/1}".

papertype := afourpaper.

:- regtype supported_papertype/1 
# "Possible papertypes: @includedef{supported_papertype/1}".

supported_papertype := letterpaper.
supported_papertype := smallbook.
supported_papertype := afourpaper.
supported_papertype := afourlatex.
supported_papertype := afourwide.
supported_papertype := afourthesis.


:- pred libtexinfo/1 => yesno

# "If set to yes the @file{texinfo.tex} file that comes with the lpdoc
   distribution will be used when generating manuals in formats such
   as @tt{dvi} and @tt{ps}. Otherwise, the texinfo file that comes
   with your @apl{tex} installation will be used. It is highly
   recommended that you leave this set to @tt{'yes'} as below:

   Example: @includedef{libtexinfo/1}".

libtexinfo := 'yes'.

:- regtype yesno/1 
# "Enumerated type: @includedef{yesno/1}".

yesno := yes|no.


% ----------------------------------------------------------------------------
% BUILD ROOT
% ----------------------------------------------------------------------------

:- pred build_root( Dir ) => atm( Dir )

# "@var{Dir} is the directory where documentation files will be
  placed. @var{Dir}/tmp will be used for temporary/cache files.".

% REVIEW?
build_root := ~atom_concat( ~package_root_dir , 'doc/' ).

% ============================================================================
% ============================================================================
% Installation options
% (You only need to change these if you will be installing the docs somewhere)
% ============================================================================
% ============================================================================

:- use_module( library( 'distutils/package_generator' ) ).

% ----------------------------------------------------------------------------
%
% :- comment(title,"Configuration File for Package Generation").
%
% :- comment(module,"This is a sample configuration file for
%    @apl{lpdoc}.  The defaults listed are typically suggestions and/or
%    the ones used for local installation in the CLIP group machines.
%    These settings should be changed to suit your application.").
%
% :- comment(author,"David Trallero Mena").
%
% ----------------------------------------------------------------------------
%
% :- comment(filetype,user).

% ----------------------------------------------------------------------------
% EXTRA VARIABLES FOR TMPL WEBSITE
% ----------------------------------------------------------------------------

:- pred extra_tmpl_info( Vars )
	: list( Vars , tmpl_vars )

# "@var{Vars} is a list of terms that will be passed to template
website generator.".

:- regtype tmpl_vars/1.

tmpl_vars( Name = Value ) :- 
	atm( Name  ), 
	atm( Value ).

extra_tmpl_info( [] ).

%-----------------------------------------------------------------------------
% WHERE MANUALS WILL BE INSTALLED
%-----------------------------------------------------------------------------

:- pred htmldir( Dir ) => atm( Dir )

# "Directory where the html manual will be installed.".

% REVIEW?
htmldir := ~atom_concat( ['/home/', ~get_pwnam, '/public_html/lpdoc_docs/']).

% Usual value for src instalation
% htmldir    := ~atom_concat( ~package_root_dir , 'doc/' ).


:- pred htmlurl( Dir ) => atm( Dir )

# "@var{Url} is the url path to access to generated website from a
  browser.

  Example: @includedef{htmlurl/1}".

htmlurl    := ~atom_concat( ['/~', ~get_pwnam, '/lpdoc_docs/']).

% Usual value for src instalation
% htmlurl := ~htmldir.


:- pred htmlstyle/1 => filename

# "Define this to point to a style sheet to be used for any html pages
   generated (a copy of the style sheet will be made). Set it to
   @tt{no} to avoid using any style sheet. 

   Example: htmlstyle := ~atom_concat([~lpdoclib,'/default.css'])". 
   % @includedef{htmlstyle/1}".

htmlstyle := no.


:- pred docdir/1 => dirpath

# "Define this to be the dir in which you want the document(s) installed. 

   Example: @includedef{docdir/1}".

% REVIEW?
docdir := ~htmldir.


:- pred infodir/1 => dirpath # "Define this to be the dir in which you
   want the info file installed".

infodir := ~htmldir.


:- pred mandir/1 => dirpath # "This is the directory in which the man
   file will be installed.".

mandir := ~htmldir.


%-----------------------------------------------------------------------------

:- pred perms( ExecPermissions ) => permission_term

# "Define this to be the mode for automatically generated
   directories and data or executable files.

   Example: @includedef{perms/1}".

perms := perm(rwX, rwX, rX).


:- regtype permission_term/1 # "Permisions: @includedef{permission_term/1}.".

permission_term(perm(User,Group,Others)) :-
	permission(User),
	permission(Group),
	permission(Others).

:- regtype permission(P) # "Possible permisions:
                            @includedef{permission/1}.".

permission := x.
permission := w.
permission := wx.
permission := r.
permission := rx.
permission := rX.
permission := rw.
permission := rwx.
permission := 'X'.
permission := wX.
permission := rwX.


:- pred owner( Owner ) => atom

# "Define this to be the user for automatically generated directories
   and files.

   Example: @includedef{owner/1}".

% REVIEW
owner := ~get_pwnam.

:- pred group( Owner ) => atom

# "Define this to be the group for automatically generated directories
   and files.

   Example: @includedef{group/1}".

% REVIEW
group := ~get_grnam.


wwwhtpasswd := '/usr/local/etc/httpd/support/htpasswd'.


% OBSOLETE??
%
% :- pred compresscommand/1 => filename
%
% # "If you would like files containing the manuals (@tt{.dvi},
%    @tt{.ps}, etc.) to be compressed on installation, set this to
%    the command to be used for such compression.
%
%    Example: @includedef{compresscommand/1}".
%
% compresscommand := 'gzip -f'.
%
%
% :- pred compressext/1 => atm
%
% # "If you have defined a compression command, define this to be the
%   suffix (extension) given by this command to compressed files.
%
%    Example: @includedef{compressext/1}".
%
% compressext := 'gz'.

% ----------------------------------------------------------------------------
% The settings below are important only for generating html and info *indices*
% ----------------------------------------------------------------------------

:- pred infodir_headfile/1 => filename

# "Define this to be file containing the header for the @apl{info}
   '@tt{dir}' index.  '@tt{dir}' entries generated by @apl{lpdoc}
   are placed in a @tt{dir} file that starts with
   @pred{infodir_headfile/1} and finishes with
   @pred{infodir_tailfile/1}.

   Example: @includedef{infodir_headfile/1}".

infodir_headfile := ~atom_concat([~component_src( lpdoc ),'/lib/Head_clip.info']).

:- pred infodir_tailfile/1 => filename

# "Define this to be file containing tail for the for the @apl{info}
   '@tt{dir}' index.  '@tt{dir}' entries generated by @apl{lpdoc}
   are placed in a @tt{dir} file that starts with
   @pred{infodir_headfile/1} and finishes with
   @pred{infodir_tailfile/1}.

   Example: @includedef{infodir_tailfile/1}".

infodir_tailfile := ~atom_concat([~component_src( lpdoc ),'/lib/Tail_clip.info']).

% End of SETTINGS
