:- module(dbserver, 
	[
	    db_handler/3, 
	    db_maintenance/3, 
	    dying/2,
	    die/1
	],
	[
	    assertions, basicmodes, regtypes, runtime_ops, 
	    pillow, persdb, persdb_mysql
	]).

:- comment(title,"An active module based dbserver with interface to
                  SQL remote DBs").

:- comment(author,"Jose Manuel Gomez Perez").

%%%%%% A server (an active module) to manage persistent databases %%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%% Needed modules %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:- use_module(engine(internals), [initialization/1]).
:- use_module(library(lists)).
:- use_module(library(system)).
:- use_module(library(aggregates)).
:- use_module(library(dynamic)).
:- use_module(library(streams)).
:- use_module(library(read)).
:- use_module(library(write)).
:- use_module(library(terms), [atom_concat/2]).
:- use_module(library(patterns)).
:- use_module(library(errhandle)).

:- use_module('lib/html_generation').
:- use_module('lib/html_templates_handle').
:- use_module('lib/html_format_handle').
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%% Database description  and html format predicates %%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% All of these data predicates correlate with files of the same name with 
% extension .pl stored in directory DBTYPES. All of them are used to 
% configure a certain aspect of the system.

% Moved to html_generation.pl
% :- data html_format/5.                       % Defines html specific types
:- data html2sql/2.                          % Defines translation of html
                                             % types into canonical SQL types
:- data sql2html/3.                          % Viceversa
:- data option_table_size/2.                 % Defines sizes of input forms
:- data dbconn_desc/6.                       % Describes DBs management itf
:- data search_expire/1.                     % Searches' date of caducity
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Set of static persistent predicates used by dbserver which are physically 
% stored in directory DBDESC.
% text_option/3                              Stores system to user messages
% search/3                                   Stores user search results
% dbtables_desc/3                            Stores table descriptions
% sql_location/2                             Stores remote SQL DBs locs
% table_stats/6                              Stores system statistics
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Auxiliary data predicates used in run time:

:- data dbtable/5.                   % ID, Description, Arity, FieldTypes, Site
                                     % Correlates to persistent predicate
                                     % dbtables_desc/3

% Moved to html_template_handle.pl
% :- data html_option/4.               % Name, Explanation, Params, HTML_code
                                     % Correlates to the system html templates
                                     % stored in DBTEMPLATES directory

:- data db_dir/1, db_url/1, db_types/1, db_templates/1. % End in '/' !
:- data db_os/1.
                                     % Index the location of several resources 
                                     % of the system

:- data launch_date/1.               % date of dbserver last launch

:- data sql_persistent_location/2.   % Description of remote SQL DBs.
                                     % Correlates to persistent predicate 
                                     % sql_location/2
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- include(password).                % Defines fact password/1

%%%%%%%%%%%%%%%%%%%%%%%%%%% Basic doc types %%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- comment(doinclude, passwd/1).
:- regtype passwd(X) # "@var{X} is an identifier for the client type.".

:- comment(doinclude, dic/1).
:- regtype dic(X) # "@var{X} is a Pillow dictionary.".

dic(_).

:- comment(doinclude, output_page/1).
:- regtype output_page(X) # "@var{X} is an HTML page.".

output_page(_).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%% System services implementation %%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- comment(usage,"This module is typically compiled as an
   @index{active module} (i.e., a @index{server}). Clients can access
   this server by using the active modules usage convention:
@begin{verbatim}
:- use_active_module(dbserver).
@end{verbatim}
").

:- pred db_handler(+Input, +Passwd, -Response) :: dic * passwd * output_page
	
   # "Used to provide an interface to the service requested by the generic 
      client. Once this predicate identifies the required service, it calls 
      the particular predicate that implements it. ".

db_handler(Input, Passwd, Response):-
	get_form_value(Input, service, Value),
	call_service(Value, Input, Passwd, Response).

:- pred call_service(+Service, +Input, +Passwd, -Response) 

   #  "Calls the required service @var{Service} and obtains its results in 
       @var{Response}. The other input parameters are used by the predicate 
       that implements the service required. ".

call_service(system_status_service,_,_,Response):- !,
	status(Response).
call_service(db_connect_service, Input, Passwd, Response):- !,
	connect_to_DB(Input, Passwd, Response).
call_service(select_query_modify_service, _Input, Passwd, Response):- !,
	get_relationsdb(Passwd, Response).
call_service(query_modify_service, Input, Passwd, Response):- !,
	consultdb(Input, Passwd, Response).
call_service(complex_query_service, Input, Passwd, Response):- !,
	complex_query(Input, Passwd, Response).
call_service(modify_options_service, Input, Passwd, Response):- !,
	consultdboptions(Input, Passwd, Response).
call_service(create_table_service, Input, Passwd, Response):- !,
	create_destroy_table(Input,Passwd, Response).
call_service(select_table_to_edit_service, _Input, Passwd, Response):- !,
	select_table_to_edit(Passwd, Response).
call_service(table_to_edit_service, Input, Passwd, Response):- !,
	edit_table(Input, Passwd, Response).
call_service(reload_templates_service, _Input, Passwd, Response):- !,
	updatetemplates(Passwd, Response).

:-pred db_maintenance(+Input, +Passwd, -Response) :: dic * passwd * output_page

   #  "According to the kind of user that calls this predicate, the system 
       generates an interface which provides access to the whole of its 
       services (in the administrator case) or to a restricted set of them 
       (in the case of a common user). Typically used in installation time to
       automatically generate both common user and administrator front pages.".

db_maintenance(Input, Passwd, Response):-
	get_form_value(Input, maintenance, yes),!,
	db_handler(Input, Passwd, Response).
	
db_maintenance(_Input, Passwd, Response):-	
	html_option(common_footer, [], Footer),
	html_option(main_page_heading, [Heading], Head),
	html_option(page_body, [[Head, Content, Footer]], Body),
	html_option(admin_content, [Links], Content),
	html_option(ok_button,[],Ok_Button),
	set_actions(service, Passwd, "tckxplyv", Services),
% If variable Passwd is not ground the user is considered to be a common one
	(var(Passwd) ->
	 platform_cgi('webdb_front', CGI),
%	 CGI='webdb_front.cgi',
	 text_option(webdb_heading,_, Heading),
	 text_option(webdb_title,_, Title)
	;
%	 CGI='admin/db_maintenance.cgi',
	 platform_cgi('admin/db_maintenance', CGI),
	 text_option(maintenance_title,_, Title),
	 text_option(maintenance_heading,_, Heading)),
	 Links = [
		     start_form(CGI),
		     Services,
		     Ok_Button,
		     input(hidden, [name=maintenance, value=yes]),
		     end_form
		 ],
        full_reply_page(Title, Body, Response).

:- pred connect_to_DB(+Input, +Passwd, -Response) :: dic * passwd * output_page

   # "Implements remote SQL database connection management service. 
      Provides an interface used to connect to a given remote SQL database 
      through ODBC given a user password and an IP address. It can be used to
      unlink the system from a previously connected database too. ".

connect_to_DB(Input, Passwd, Response):-
	check_passwd(Passwd),
	get_one_pred_form_vals(Input, dbconn_desc, 6, LFields, LValues),
	operate_on_dbs(Input, LValues, Reply),
	html_option(admin_footer, [], Footer),
	text_option(db_connection_title,_, Title),
	text_option(db_connection_heading,_,Heading),
	html_option(page_body, [[Head, Content, Footer]], Body),
	html_option(admin_heading, [Heading], Head),
	html_option(admin_content, [DBManagement], Content),        
	html_option(ok_button,[],Ok_Button),
	html_option(reset_button,[],Reset_Button),
	html_option(form_buttons,[Ok_Button, Reset_Button], Buttons),
	text_option(add_database_message, _, AText),
	html_option(general_add, [AText], Add_db_message),
	dbconn_desc(FD1,FD2,FD3,FD4,FD5,FD6),
	form_(LFields, [FD1,FD2,FD3,FD4,FD5,FD6], InputFields),
	html_option(general_table,['2',InputFields], Form),
	findall(Site, sql_persistent_location(Site, _), SitesList), 
	(SitesList=[] ->
	 Del_db_message=[],
	 Del_Menu=[],
	 Message=[],
	 Current_DBs=[]	
	;
	 findall(d(DBId, DBHandle, User, IP_Add, Port),	sql_persistent_location(DBId, db(DBHandle, User, _, IP_Add:Port)), DBList),
	 text_option(del_database_message, _, DDText),
	 html_option(big_bold_message, [DDText], Del_db_message),
	 db_options(SitesList, Options),
	 html_option(db_connection_message,[],Message),
	 get_descriptions([FD1,FD2,FD3,FD5,FD6], Descs),
	 name_list_to_table_heading(Descs, DBTheading),
	 get_general_rows(DBList, DBRows),
	 data_to_table_row(DBTheading, Row1),
	 Rows=[Row1|DBRows],
	 html_option(general_table, ['5', Rows], Current_DBs),
	 Del_Menu=select([name=dest_list], [option([value=not_destroy], 'Keep current links')|Options])),
	 platform_cgi('admin/db_generic_client', CGI),
	 DBManagement=[
			  Reply,
			  Message,
			  start_form(CGI),
			  Current_DBs,
			  Del_db_message,
			  Del_Menu,
			  Add_db_message,
			  Form,
			  Buttons,
			  input(hidden, [name=selected, value=yes]),
			  input(hidden, [name=service,value=db_connect_service]),
			  end_form
                      ],
        full_reply_page(Title, Body, Response).	

:- pred get_descriptions(+FieldTypes, -FieldDescs)

   # " For a given list @var{FieldTypes} of html types in dbtables_desc/3 
       format obtains a list @var{FieldDescs} that contains the descriptive 
       names of any. ".

get_descriptions([], []).
get_descriptions([FT1|FTs],[Desc1|Descs]):-
	type_info(FT1, _, Desc1),
	get_descriptions(FTs, Descs).

:- pred operate_on_dbs(+Input, +Lvalues, -Reply)

   # " Generates a form that takes from the user the information the system 
       needs to connect to a given remote SQL database or unlink it from an 
       already linked one. ".

operate_on_dbs(Input, LValues, Reply):-
	get_form_value(Input, selected, yes), !,
	get_form_value(Input,dest_list,DB2Dest),
	unlink_DB(DB2Dest),
	turn2terms(LValues, LTerms),
	DBLinks_pred=..[c_form|LTerms],
	(ground(DBLinks_pred)->
	 turn2loc_syntax(DBLinks_pred, Loc),
	 link_DB(Loc, Add_Reply)
	;
	 text_option(insert_warning, _, Message),
	 html_option(consult_table_message, [Message], Add_Reply)
	),
	(DB2Dest=not_destroy ->
	 Reply=Add_Reply
	;
	 Reply=[]
	).
operate_on_dbs(_,_,[]).

:- pred turn2loc_syntax(+InFormat, -OutFormat)

   # "Turns some data into the @pred{sql_persistent_location/2} particular
     format. ".    

turn2loc_syntax(c_form(DbId, DbH, U, Pass, IP, Port), 
	sql_persistent_location(DbId, db(DbH, U, Pass, IP:Port))).
	
:- pred unlink_DB(+Link2Destroy)
   
   # "When the user has selected which remote SQL Database is to be unlinked 
      from the system, carries out the particular protocol necessary to do it 
      cleanly. ".

unlink_DB(not_destroy):- !.
unlink_DB(DbId):-
	reset_stats(DbId),
	pretract_all(dbtables_desc(_, DbId,_)),
	pretract_all(sql_location(DbId,_)),
	retractall(dbtable(_,_,_,DbId,_)),
	retractall(sql_persistent_location(DbId,_)).

:- pred link_DB(+Database2Link, -Reply)

   # "Tests if the remote SQL database is already linked from the same source. 
      If not, links it to he system. ".

link_DB(NewLoc, Reply):-
	call(NewLoc), !,
	text_option(repeated_elem, _, Message),
	html_option(consult_table_message, [Message], Reply).
link_DB(sql_persistent_location(DbId, db(DbH, U, _, _:_)), Reply):-
	(sql_persistent_location(DbId, db(_, U, _, _:_));
	 sql_persistent_location(_, db(DbH, U, _, _:_))), !,
	text_option(repeated_Id, _, Message),
	html_option(consult_table_message, [Message], Reply).
link_DB(sql_persistent_location(DbId, DBHandle), []):-
	data_facts:assertz_fact(sql_persistent_location(DbId, DBHandle)),
	persdbrt:assertz_fact(sql_location(DbId, DBHandle)),
	sql_get_tables(DbId, TablesList),
	include_tables(DbId, TablesList).

:- pred include_tables(+DatabaseId, -Reply)

   # "Includes the tables of a remote SQL database into the system's database 
      tables description. ".

include_tables(_,[]).
include_tables(DbId,[T1|Ts]):-
	sql_table_types(DbId, T1, Double_list), 
	obtain_ids_types(Double_list, A_list, T_list),	
	SQLDef=..[T1|A_list],
	PrologDef=..[T1|T_list],
	make_sql_persistent(PrologDef, SQLDef, DbId),
	attach2DB(DbId, T1, A_list, T_list),
	include_tables(DbId,Ts).

:- pred obtain_ids_types(+Pairs, -Ids, -Types)

   # "Transforms a list of pairs (FieldIdentifier, FieldSQLtype) in a list of 
      FieldIdentifier and a list of FieldSQLtype." .

obtain_ids_types([],[],[]).
obtain_ids_types([[Id, Type]|Rest], [Id|Ids], [Type|Ts]):-
	obtain_ids_types(Rest, Ids, Ts).

:- pred attach2DB(+DbId, +T1, +A_list, +T_list)

   # "Given a remote SQL database identifier @var{DbId}, a table name of that
      database @var{T1}, its attributes identifier list @var{A_list} and its
      attributes type list @var{T_lists}, attaches that particular table to 
      the system. ".

attach2DB(DbId, T1, A_list, T_list):-
	get_html_args(A_list, T_list, Html_args),
	NewTableDesc=..[T1|Html_args],
	date(Date),	
	persdbrt:assertz_fact(dbtables_desc(NewTableDesc, DbId, Date)),
	atom_concat([T1,'_',DbId], TName),
	length(A_list, A),
	data_facts:assertz_fact(dbtable(TName, A, NewTableDesc, DbId, Date)),
	initialize_stat(TName).
	
:- pred get_html_args(+A_list, +T_list, -Html_args_list)

   # " Given a list of attributes @var{A_list} and a correlated list of their 
       SQL types @var{T_list}, obtains a list of @pred{dbtables_desc/3} syntax
       html attributes. ".

get_html_args([], [], []).
get_html_args([A|A_list], [T|T_list], [H|Html_args]):-
	sql2html_aux(T, A, H),
	get_html_args(A_list, T_list, Html_args).

:- pred sql2html_aux(+SQLtype, +AttributeId, +HTMLarg)

   # "Given an SQL type @var{SQLtype} and an attribute identifier 
      @var{AttributeId}, obtains a @pred{dbtables_desc/3} syntax
       html attribute @var{HTMLarg}. Default is a name type".

sql2html_aux(T, A, H):-
	sql2html(T, A, H), !.
sql2html_aux(_, A, name(A,'',20)).

:- comment(get_relationsdb/2,"Implements with @pred{consultdb/3} the query 
      one tuple service. It displays an HTML form with a menu where
      the user can choose the table to operate on. It makes out wether
      the user is an authorised client or a common one by checking the argument
      @var{ClientType}. If it's an authorised client, the user will be
      allowed to insert, delete or modify data, otherwise just consult
      the system to make standard search operations.").

:- pred get_relationsdb(+ClientType, -OutputFields) :: passwd * output_page

   # "@var{OutputFields} is a form where the user can choose a
      relation to operate on (the operations permitted depend on
      @var{ClientType}).".

get_relationsdb(ClientType, Reply):-
       get_opts_clienttype(ClientType, CGI, Footer),
       text_option(choose_table_title,_, Title),
       text_option(choose_table_heading,_,Heading),
       html_option(page_body, [[Head, Content, Footer]], Body),
       html_option(admin_heading, [Heading], Head),
       html_option(admin_content, [Choose_table], Content),
       text_option(choose_table_message,_,CTMText),
       html_option(general_choose, [CTMText],Message),
       html_option(ok_button,[],Ok_Button),
       findall(TI, dbtable(TI, _, _, _,_), TIList),
       table_options(TIList, Options),
       Choose_table = [
			  Message,
			  start_form(CGI),
			  select([name=table_name], Options),
			  input(hidden,[name=action, value=empty_form]),
			  input(hidden,[name=searchmode, value=normal]),
			  input(hidden,[name=service,value=query_modify_service]),
			  \\, \\,
			  Ok_Button,
			  end_form
		        ],
        full_reply_page(Title, Body, Reply).

:- pred get_opts_clienttype(+Password, -CGI, -Footer) 

   # "Sets the CGI client to interact with the user depending on the password 
      produced. If no password is given, a simple CGI client with no 
      privilegies is prompted. In other case, if the password is right the CGI 
      will be one with full access to the system services, otherwise the 
      system acts against the aggression. ".

get_opts_clienttype(V, CGI, Footer) :-
	var(V), !,
	platform_cgi('webdb_client', CGI),
	html_option(common_footer, [], Footer).
get_opts_clienttype(Passwd, CGI, Footer) :-
	password(Passwd), !,
	platform_cgi('admin/db_generic_client', CGI),
	html_option(admin_footer, [], Footer).
get_opts_clienttype(Bad, _, _) :-
	bad_passwd(Bad).

:- pred table_options(+TablesList, -Menu)
	
   # "Given a list of tables @var{TableList}, generates a Pillow menu in order
      to allow the user to choose one @var{Menu}. ".

table_options([], []).
table_options([TI|TIList], [Opt|OptList]) :-
	dbtable(TI, _, _, _, _),
	Opt = option([value=TI], TI),
	table_options(TIList, OptList).

:- pred db_options(+DatabaseList, -Menu)
	
   # "Given a list of databases @var{DatabaseList}, generates a Pillow menu 
      in order to allow the user to choose one @var{Menu}. ".

db_options([], []).
db_options([Site|SiteList], [Opt|OptList]) :-
        dbtable(_, _, _, Site,_),
        Opt = option([value=Site], Site),
        db_options(SiteList, OptList).


:- comment(consultdb/3,"Once the user has decided on which table to
   operate on through the service provided by
   @pred{get_relationsdb/2}, @pred{consultdb/3} displays the system
   consult interface. Through this interface the user can choose which
   operation to carry out over the database.").

:- pred consultdb(+Input, +ClientType, -Response) :: dic * passwd * output_page

   # "Consults @var{Input} and returns @var{Response} as a response
      for a given client type @var{ClientType}.".

consultdb(Input, ClientType, Response) :-
	get_form_value(Input, table_name, Table),!,
	dbtable(Table, Arity, _, _, _),
	Arity > 0, !,
	get_form_value(Input, action, Action), 
	(Action=refresh ->
	 get_form_value(Input, search, Session),
	 make_refresh(Session, Input, Table, Arity, Next, Reply, Data)
	;
	 get_one_pred_form_vals(Input, Table, Arity, LFields, LValues),
	 full_reply(Action, Input,  ClientType, Table, LFields, 
	            LValues, Next, Reply, Data, Success)),
	((Action=overview; (Action=full_search,var(ClientType))), Success=yes ->
	  Buttons=[],
	  Select_action=[],
	  Select_Search_Mode=[],
	  Search_Mode=[]
	;
	  html_option(ok_button, [], Ok_Button),
	  html_option(reset_button, [], Reset_Button),
	  html_option(form_buttons, [Ok_Button, Reset_Button], Buttons),
	  (((Action\==full_search, Action\==overview); Success=no) ->
	   text_option(select_action_message, _, SAText),
	   html_option(general_sel_mess, [SAText],Select_action),
	   text_option(select_search_mode_mess, _, SSMText),
	   html_option(general_sel_mess, [SSMText],Select_Search_Mode),
	   get_form_value(Input, searchmode, Search),
	   set_search_mode(Search, "ncr", Search_Mode)
	  ;
	   Select_action=[],
	   Select_Search_Mode=[],
	   Search_Mode=[])),  
	set_actions(action, ClientType, Next, NextActions),
	set_footer_CGI(ClientType, Footer, CGI),
	text_option(database_consult_title,_, Title),
	html_option(consult_table_heading, [Table], Head),
	html_option(page_body, [[Head, Content, Footer]], Body),
	html_option(admin_content, [ConsultDB], Content),
	html_option(ok_button, [], Ok_Button),
	html_option(reset_button, [], Reset_Button),
       ConsultDB = [
		       Reply,
		       start_form(CGI),
		       Data,
		       Select_action,
		       NextActions,
		       Select_Search_Mode,
		       Search_Mode,
		       input(hidden, [name=table_name, value=Table]),
		       input(hidden, [name=service, value=query_modify_service]),
		       Buttons,
		       end_form
                   ],
        full_reply_page(Title, Body, Response).
consultdb(_, _, Response):-
	text_option(database_consult_title,_, Title),
	text_option(table_not_ready_mess, _, Message),
	db_url(URL), 
	message_page(Title, Message, URL, Response).	

:- pred prepare_search(+Search_Mode, +Table, +LValues, -LPreds)

   # "Obtain the matches of a query in a table @var{Table} given the 
      search mode @var{Search_Mode} and the input values @var{LValues}. 
      Basically, transforms the input form fields and values into a prolog 
      term and finds all the terms contained in the table that match it. ".

prepare_search(normal, Table, LValues, LPreds):- !,
	dbtable(Table, _, _, SiteId,_),
	get_single_query(Table, LValues, Pred),
	itf_findall(SiteId, Pred, Pred, LPreds).

prepare_search(Search_Mode,Table, LValues, LPreds):-
	get_single_query(Table, LValues, Pred),
	dbtable(Table, Arity, _, SiteId, _),
	get_free_vars(1, Arity, VarList),
	(SiteId = local ->
	 PredAux=..[Table|VarList]
	;
	 get_remote_table_name(SiteId, Table, TName),
	 PredAux=..[TName|VarList]),
	itf_findall(SiteId, PredAux, PredAux, LPredsAux),
	make_filter(Search_Mode, Pred, LPredsAux, LPreds).

:- pred make_filter(+Search_Mode, +Pred, +LPredsIn, -LPredsOut)

   # "Given a search criterium @var{Search_Mode}, a pattern @var{Pred}, and a
      list of candidates for match @var{LPredsIn} obtains a list with do match
      according to that criterium. ".

make_filter(case_ins, Pred, LPredsAux, LPreds):-	!, 
	filter_cimatch(Pred, LPredsAux, LPreds).
make_filter(reg_exp, Pred, LPredsAux, LPreds):-
	filter_regmatch(Pred, LPredsAux, LPreds).

:- pred filter_cimatch(+Pred, +LPredsIn, -LPredsOut)

   # "Given the search criterium case_insensitive_match, a pattern @var{Pred}, 
      and a list of candidates for match @var{LPredsIn}, obtains a list with 
      do match according to that criterium. ".

filter_cimatch(_, [], []).
filter_cimatch(Pred, [PredAux1|PAs], [PredAux1|Ps]):-
	case_insensitive_match(Pred, PredAux1), !,
	filter_cimatch(Pred, PAs, Ps).
filter_cimatch(Pred, [_PredAux1|PAs], Ps):-
	filter_cimatch(Pred, PAs, Ps).

:- pred filter_regmatch(+Pred, +LPredsIn, -LPredsOut)

   # "Given the search criterium regular_expresion_match, a pattern @var{Pred},
      and a list of candidates for match @var{LPredsIn}, obtains a list with 
      do match according to that criterium. ".

filter_regmatch(_, [], []).
filter_regmatch(Pred, [PredAux1|PAs], [PredAux1|Ps]):-
	match_pattern_pred(Pred, PredAux1), !, 
	filter_regmatch(Pred, PAs, Ps).
filter_regmatch(Pred, [_PredAux1|PAs], Ps):-
	filter_regmatch(Pred, PAs, Ps).

:- pred register_search(+LPreds, -Session)

   # "Registers the list of terms resulting from a query @var{LPreds}, and 
      produces a session number @var{Session} associated to it. ".

register_search(LPreds, Session):-
	enumerate(LPreds, LNumbPreds, Length),
	time(Session),
	clear_old_sessions(Session),
	persdbrt:assertz_fact(search(Session, LNumbPreds, Length)). 

:- pred set_search_mode(+Action, +Actions, -Search_Mode)

   # "Produces a checkpoint list @var{Search_Mode} where the user can choose 
      the kind of match desired for a particular one table query. ".

set_search_mode(Action, Actions, Search_Mode):-
	Actions=[A|_],
	char_to_search_mode(A, Default),
	compute_search_mode(Action, Actions, Default, Search_Mode).

compute_search_mode(_, [], _, []).
compute_search_mode(Action, [A|As], Def, [S1|Ss]):-
	char_to_search_mode(A, S_mode),
	compute_checked(Action, S_mode, Checked),
	text_option(S_mode, _, Text),
	html_option(search_mode_display,[S_mode,Text,Checked], S1),
	compute_search_mode(Action, As, Def, Ss).

char_to_search_mode(0'n, normal).
char_to_search_mode(0'c, case_ins).
char_to_search_mode(0'r, reg_exp).

:- pred full_reply(Action, Input, Table, LFields, LValues, 
	Mode, Reply, Data, Success) 

   # "For a given table @var{Table} and fields @var{LFields} and
      values @var{LValues} of a form it gives the accurate reply
      @var{Reply}, provides a new form data @var{Data} and a new mode
      @var{Mode} for the following actions.".

% Starting a search from scratch.
full_reply(empty_form,_Input,_,Table, LFields,_LValues, "sofid", [], Form, _):-
        dbtable(Table, _, Types, _, _),
	form(LFields, Types, Form).
% Typical search action
full_reply(search, Input,_,Table,LFields, LValues, Actions, Reply, Form, yes):-
	get_form_value(Input, searchmode, Search_Mode),
	prepare_search(Search_Mode, Table, LValues, LPreds),
	LPreds \== [], !,
	register_search(LPreds, Session),
	search_vals(Session, 1, Table, LFields, Actions, Reply, Form),
	update_stats(Table, s).
full_reply(search,_Input,_,Table, LFields,_LValues, "sofid", Reply, Form, no):-
	text_option(no_found_reply, _, Message),
	html_option(consult_table_message, [Message], Reply),
        dbtable(Table, _, Types, _, _),
	form(LFields, Types, Form).
% Show all matching tuples in a table (low level) search action
full_reply(overview, Input, _,Table, _LFields, LValues, [], Reply, Form, yes):-
	get_form_value(Input, searchmode, Search_Mode),
	prepare_search(Search_Mode, Table, LValues, LPreds),
	LPreds \== [], !,
	register_search(LPreds, Session),
	overview_vals(Session, Table, Reply, Form),
	update_stats(Table, s).	
full_reply(overview,_Input,_,Table,LFields,_LValues,"sofid", Reply, Form, no):-
	text_option(no_found_reply, _, Message),
	html_option(consult_table_message, [Message], Reply),
        dbtable(Table, _, Types, _, _),
	form(LFields, Types, Form).
% Show all matching tuples in a table search action
full_reply(full_search, Input, CType,Table, _LFields, LValues, Actions, Reply, OutputForm, yes):-
	get_form_value(Input, searchmode, Search_Mode),
	prepare_search(Search_Mode, Table, LValues, LPreds),
	LPreds \== [], !,
	register_search(LPreds, Session),
	full_search(Table, Session, CType, Reply, Form),
	OutputForm=[input(hidden, [name=searchmode, value=normal])|Form],
	(var(CType) ->
	 Actions=[]
	;
	 Actions="r"),
	update_stats(Table, s).
full_reply(full_search,_Input,_,Table,LFields,_LValues,"sofid",Reply,Form,no):-
	text_option(no_found_reply, _, Message),
	html_option(consult_table_message, [Message], Reply),
        dbtable(Table, _, Types, _, _),
	form(LFields, Types, Form).
% The user asks for the next element of a search
full_reply(ahead, Input, _,Table, LFields,_LValues, Actions, Reply, Form,_):-
	get_form_value(Input, search, Session),
	get_form_value(Input, currelem, LastElem),
	Elem is LastElem+1,
	search_vals(Session, Elem, Table, LFields, Actions, Reply, Form).
% The user asks for the previous element of a search
full_reply(back, Input, _,Table, LFields,_LValues, Actions, Reply, Form,_):-
	get_form_value(Input, search, Session),
	get_form_value(Input, currelem, LastElem),
	Elem is LastElem-1,
	search_vals(Session, Elem, Table, LFields, Actions, Reply, Form).
% Insertion of a new element of the relation
full_reply(insert,_Input,_,Table, LFields, LValues, "esofid", Reply, Form, _):-
	dbtable(Table, _, Types, SiteId, _),
	get_single_query(Table, LValues, Pred),
	pred_to_form(Pred, Types, LFields, Form),
	( form_empty_list_values(LValues) ->
	   text_option(insert_warning, _, Message)
        ; itf_call(SiteId, Pred) ->
	   text_option(repeated_elem, _, Message)
        ; itf_assert(Pred), 
          text_option(element_inserted, _, Message)
       ),
       html_option(consult_table_message, [Message], Reply),
	update_stats(Table, i).
% Modification of a previously inserted element
full_reply(modify, Input, _,Table, LFields, LValues, Actions, Reply, Form,_):-
	get_single_query(Table, LValues, NPred),
	get_form_value(Input, search, Session),
	search(Session, LPreds, Length),
	get_form_value(Input, currelem, Elem),
	select(pred(Elem, Pred), LPreds, LNewPreds),
	(form_empty_list_values(LValues) ->
	 text_option(modify_warning,_, Message),
	 output_search(Table, Pred, LFields, Session, Elem, Form)
	;
	 persdbrt:retract_fact(search(Session, LPreds, Length)), 
	 itf_retract(Pred), !,
	 itf_assert(NPred),
	 persdbrt:assertz_fact(search(Session, [pred(Elem, NPred)|LNewPreds], 
	                       Length)),
	 text_option(element_modified,_, Message),
	 output_search(Table, NPred, LFields, Session, Elem, Form)),
	html_option(consult_table_message, [Message], Reply),
	next_actions(Elem, Length, Actions),
	update_stats(Table, m).
% Deletion of an element
full_reply(delete, Input, _,Table, LFields, LValues, "sofid", Reply, Form, _):-
	get_form_value(Input, deletion, yes), !,
	get_form_value(Input, searchmode, Search_Mode),
	prepare_search(Search_Mode, Table, LValues, LPreds),
	itf_retractall(LPreds), % Delete everything !
        dbtable(Table, _, Types, _, _),
	form(LFields, Types, Form),
	text_option(deletion_complete,_, Message),
	html_option(consult_table_message, [Message], Reply),
	update_stats(Table, d).
full_reply(delete,_Input, _,Table, LFields, LValues, "dsofi", Reply, Form, _):-
	get_single_query(Table, LValues, Pred),
	dbtable(Table, _, Types, _, _),
	pred_to_form(Pred, Types, LFields, NF),
	Form=[input(hidden, [name=deletion, value=yes])|NF],
	html_option(deletion_warning, [], Message),
	html_option(consult_table_message, [Message], Reply).
% Search expired. Its contents will no longer be kept.
full_reply(_Command,_Input,_,Table,LFields,_LValues, "sofid", Reply, Form, _):-
        dbtable(Table, _, Types, _, _),
	form(LFields, Types, Form),
	text_option(search_expired, _, Message),
	html_option(consult_table_message, [Message], Reply).

:- pred search_vals(+Session, +Elem, +Table, +LFields, -Actions, -Reply, -Form)
   # "After having made a one table search, this predicate generates a new 
      form @var{form} that will allow a new set of actions @var{Actions} 
      according to the circumstances of the previous search, as well as an 
      appropiate reply @var{Reply}. ".

search_vals(Session, Elem, Table, LFields, Actions, Reply, Form) :-
        search(Session, LPreds, Length),
        ( (Elem > Length ; Elem < 1) ->
            text_option(no_found_reply, _, Message),
            html_option(consult_table_message, [Message], Reply),
	    dbtable(Table, _, Types, _, _),
            form(LFields, Types, Form),
            Actions = "sofid"
        ; 
          member(pred(Elem, Pred), LPreds),
          output_search(Table, Pred, LFields, Session, Elem, Form),
          number_codes(Elem,ES), atom_codes(AElem,ES),
          number_codes(Length,LS), atom_codes(ALength,LS),
          html_option(match_message, [AElem,ALength], Reply),
          next_actions(Elem, Length, Actions)
        ).

:- pred make_refresh(+Session, +Input, +Table, +Arity, -NewActions, -Reply, -Data)

   # "After having made a full_search/5 operation this is commonly used to 
      update the tuples which may have been edited in such operation. It also
      generates a new form for a new table query".

make_refresh(Session, Input, Table, Arity, "sofid", Reply, Data):-
	search(Session, LPreds, Length),
	get_n_preds_form_vals(Input, Table, Arity, Length, LLValues),
	make_changes(LLValues, LPreds, Table, NewSessionVals),
	persdbrt:retract_fact(search(Session, LPreds, Length)),
	persdbrt:assertz_fact(search(Session, NewSessionVals, Length)),
	functor_to_field_names(Table, Arity, LFields),
        dbtable(Table, _, Types, _, _),
	form(LFields, Types, Data),
	html_option(admin_short_message, ['Refreshment Completed. Make your next operation'], Reply),
	update_stats(Table, m).

:- pred make_changes(+LVals, +Lpreds, +Table, -NewPreds)

   # "Retracts from the database the old data @var{LPreds} and updates it 
      with the new data @var{NewPreds} obtained from a list of values 
      @var{LVals}. ".

make_changes([], [], _, []).	
make_changes([LVals|LLVs], [pred(N,Pred)|Ps], Table, [pred(N, NewPred)|NPs]):-
	get_single_query(Table, LVals, NewPred),
	((form_empty_list_values(LVals)) ->
	 itf_retract(Pred)
	;
	 itf_retract(Pred),
 	 itf_assert(NewPred)),
	make_changes(LLVs, Ps, Table, NPs).

:- pred next_actions(+CurrentItem, +LastItem, -Actions)

   # " Determines in a one table query search which actions @var{Actions} will
       be allowed next according with the order of the current element 
       shown. ".

next_actions(1, 1, "sofeidm") :- !.
next_actions(1, _, "asofeidm") :- !.
next_actions(L, L, "sofbeidm") :- !.
next_actions(_, _, "absofeidm").

:- pred set_actions(+Type, +ClientType, +ActionList, -Actions)

   # "Produces a checkpoint list @var{Actions} where the user can choose 
      the subsequent action. ".

set_actions(_,_, [], []).
set_actions(Type, ClientType, ActionList, Actions) :-
        ActionList = [C|_],
        char_to_action(C, Default),
        allowed_actions(ClientType, Acts),
        compute_actions(Type, Acts, ActionList, Default, Actions).

compute_actions(_,[], _, _, []).
compute_actions(Type, [C|Cs], AList, Def, [A|As]) :-
        member(C, AList), !,
        char_to_action(C, Act),
        compute_checked(Act,Def,Checked),
        text_option(Act,_,Text),
	(Type=action ->
	 Display_Template=action_display
	;
	 Display_Template=service_display),
        html_option(Display_Template, [Act, Text, Checked], A),
        compute_actions(Type, Cs, AList, Def, As).
compute_actions(Type, [_|Cs], AList, Def, As) :-
        compute_actions(Type, Cs, AList, Def, As).

compute_checked(Def,Def,checked) :- !.
compute_checked(_, _, nochecked).

%Query actions
char_to_action(0'a, ahead).
char_to_action(0'b, back).
char_to_action(0'e, empty_form).
char_to_action(0's, search).
char_to_action(0'o, overview).
char_to_action(0'f, full_search).
char_to_action(0'r, refresh).
char_to_action(0'i, insert).
char_to_action(0'd, delete).
char_to_action(0'm, modify).
%Service actions
char_to_action(0't, system_status_service). 
char_to_action(0'c, db_connect_service). 
char_to_action(0'k, select_query_modify_service).
char_to_action(0'q, query_modify_service).
char_to_action(0'x, complex_query_service). 
char_to_action(0'p, modify_options_service). 
char_to_action(0'l, create_table_service). 
char_to_action(0'y, select_table_to_edit_service). 
char_to_action(0'n, table_to_edit_service). 
char_to_action(0'v, reload_templates_service). 

allowed_actions(V,      "abesftkqx") :- var(V), !.
allowed_actions(Passwd, "abesofridmtckqxplynv") :-
        password(Passwd), !.
allowed_actions(Bad, _) :-
        bad_passwd(Bad).

:- pred overview_vals(+Session, +Table, -Reply, -Page)

   # "Generates an HTML table which displays the result of a one table search
      if an overview action was choosen. ".

overview_vals(Session, Table, Reply, Page):-
	search(Session, LPreds, Length),
	dbtable(Table,_,TableFieldsDesc,_,_),
	TableFieldsDesc=..[_|LFieldsDesc],
	number_codes(Length,LS), atom_codes(ALength,LS),
	atom_concat('Overview search resulted in ', ALength, MessAux),
	atom_concat(MessAux, ' matches', Message),
	html_option(admin_short_message, [Message], Reply),
	preds_to_data_table(LPreds, LFieldsDesc, Page).

:- pred full_search(+Table, +Session, +Passwd, -Reply, -Page)

   # "Generates an HTML table which displays the result of a one table search
      if a full search action was choosen. It also allows editing these 
      results. ".

full_search(Table, Session, CType, Reply, Page):-
	search(Session, LPreds, Length),
	extract_preds(LPreds, Preds), 
	dbtable(Table, _,TableFieldsDesc,_,_),
	TableFieldsDesc=..[_|LFieldsDesc],
	number_codes(Length,LS), atom_codes(ALength,LS),
	atom_concat('Resulting tuples: ', ALength, Message),
	html_option(admin_short_message, [Message], Reply),
	full_search_page(Session, CType, Preds, LFieldsDesc, Page).

full_search_page(Session, CType, LPreds, LFieldsDesc, Page):-
	(var(CType)-> 
	 DisplayMode = private
	;
	 DisplayMode = public),
	preds_to_form(DisplayMode, LPreds, LFieldsDesc, P),
	Page=[input(hidden, [name=search, value=Session])|P].

%%%%%%%%%%%% pred_to_form and on was here

:- pred set_footer_CGI(+Password, -Footer, -CGI) 

   # "Sets the CGI client to interact with the user and the right footer 
      depending on the password produced. ".

set_footer_CGI(V, Footer, CGI):- var(V), !,
	html_option(common_footer,[], Footer),
	platform_cgi('webdb_client', CGI).
set_footer_CGI(Passwd, Footer, CGI):-
       password(Passwd), !,
       html_option(admin_footer, [], Footer),
	platform_cgi('admin/db_generic_client', CGI).
set_footer_CGI(Bad, _, _) :-
	bad_passwd(Bad).

:- pred full_reply_page(+Title, +Body, -OutputPage)

   # "Given a title @var{Title} and a body @var{Body}, generates an HTML 
      page @var{OutputPage} with a fix structure. ".

full_reply_page(Title, Body, OutputPage) :-
	db_url(BaseURL), 
	compose_page(Title, Body, BaseURL, OutputPage).

% % Moved to html_generation.pl
% compose_page(Title, Body, BaseURL, [cgi_reply,
%                                     start,
% 				    base$[href=BaseURL],
% 				    title(Title),
% 				    Body,
% 				    end]).


:- pred enumerate(+LPreds, -LNumPreds, -Length) 

   # "Creates a list @var{LnumPreds} of predicates @tt{pred(Number,
      Pred)} where @tt{Number} gives the order of @tt{Pred} in the new
      list given a previous list of predicates
      @var{LPreds}. @var{Length} is the length of the list.".

enumerate(LPreds, LNumPreds, Length):-
	enumerate_(LPreds, 0, LNumPreds, Length).

enumerate_([P|LPreds], C, [pred(Counter,P)|LNumbPreds], Lentgh):-
	Counter is C+1,
	enumerate_(LPreds, Counter, LNumbPreds, Lentgh).
enumerate_([], C ,[], C).

	
:- pred get_single_query(+Table, +LFieldValues, -Pred) 

   # "Gets a Prolog query @var{Pred} given a relation name @var{Table}
      and its values @var{LFieldValues}.".

get_single_query(Table, LFieldValues, Pred):-
	turn2terms(LFieldValues, LFieldTerms),
	dbtable(Table, _,_,SiteId, _),
	(SiteId = local ->
	 Pred=..[Table|LFieldTerms]
	;
	 get_remote_table_name(SiteId, Table, Functor),
	 Pred=..[Functor|LFieldTerms]
	).

get_remote_table_name(SiteId, Table, TableName):-
	atom_concat('_', SiteId, SSiteId),
	atom_concat(TableName, SSiteId, Table).

:- pred turn2terms(+Values, -Terms) 

   # "Converts a List of HTML values @var{Values} into a list of
      Prolog Terms @var{Terms}.".

turn2terms([], []).
turn2terms([Value|Values], [Term|Terms]):-
	(form_empty_value(Value)->
	  var(Term)               %just declarative
        ; 
	  Term=Value),
	turn2terms(Values, Terms).

:- pred output_search(+Table, +Pred, +LFields, +Session, +Elem, -Form) 

   # "Prepares the next page for an user to keep interacting with the system 
      through the one table query interface. ".

output_search(Table, NPred, LFields, Session, Elem, Form) :-
       dbtable(Table, _, Types, _, _),
       pred_to_form(NPred, Types, LFields, F),
       Form = [input(hidden, [name=search, value=Session]),
               input(hidden, [name=currelem, value=Elem])
               |F].

:- pred clear_old_sessions(Session) # "Cleans the system of obsolete
   sessions. It eliminates every previous session to Session which is
   older than Elapsed seconds.".

clear_old_sessions(Session):-
	search_expire(Elapsed),
	PrevSession is Session - Elapsed,
	search(S, _, _), S < PrevSession,
	persdbrt:retract_fact(search(S, _, _)),
	fail.
clear_old_sessions(_).

:- pred complex_query(+Input, +ClientType, -Response) 
   :: dic * passwd * output_page

   # "Implements the SQL/Prolog Syntax search service. Provides an 
      interface which allows querying the database in both
      Prolog-like and SQL syntax. This way, the user can operate on the
      database just as if from the Prolog top level or from the SQL
      command line. ".

complex_query(Input,Passwd, Response) :-
	get_form_value(Input, query_began, yes), !,
	process_query(Input, Reply, Page),
	set_footer_CGI(Passwd, _Footer, _),
	text_option(complex_query_title,_, Title),
	text_option(complex_query_response,_, ResponseH),
	html_option(admin_heading, [ResponseH], Head),
	html_option(page_body, [[Head, Content, []]], Body),
	html_option(admin_content, [ComplexqDB], Content),
	ComplexqDB = [
			 Reply,
			 Page
		     ],
       full_reply_page(Title, Body, Response).	

complex_query(_Input,Passwd, Response) :-
	findall(table_info(Table_Name, Arity, TFD, Site), 
	        dbtable(Table_Name, Arity, TFD, Site, _), TList),
       display_tables_info(TList, TInfo),
       html_option(general_table,[2,[TInfo]], TInfoHTML),
       html_option(complex_query_instructions, [], Instructions), 
       text_option(complex_query_head, _, CQHead),
       html_option(general_heading, [CQHead], Mess),
       text_option(complex_query_title,_, Title),
       findall(Loc, sql_persistent_location(Loc, _), Locations),
       tables_locations([local|Locations], Options),
       html_option(admin_heading, [Title], Head),
       html_option(admin_content, [Complex_table], Content),
       text_option(choose_database_loc_mess, _, CDText),
       html_option(general_choose,[CDText], Loc_Prompt),
       set_footer_CGI(Passwd, Footer, CGI),
       html_option(ok_button, [], Ok_Button),
       html_option(reset_button, [], Reset_Button),
       html_option(form_buttons, [Ok_Button, Reset_Button], Buttons),
       Complex_table = [
			   Mess,
			   TInfoHTML,
			   start_form(CGI),
			   Loc_Prompt,
			   select([name=database_name], Options),
			   Instructions,
			   textinput(query, [rows=10, cols=50], 
			             "Please, write here your query"), 
			   input(hidden, [name=query_began, value=yes]),
			   input(hidden, [name=service,
			                  value=complex_query_service]),
                         Buttons,
                         end_form
		        ],
	html_option(page_body, [[Head, Content, Footer]], Body),
	full_reply_page(Title, Body, Response).

:- pred process_query(+Input, +Reply, -Response)

   # "Analizes the query written by the client in Prolog or SQL syntax and 
      generates a table @var{Response} with the variables involved and 
      their matching values. ".

process_query(Input, Reply, Response):-
	get_form_value(Input, query, AtomQuery),
	get_form_value(Input, database_name, DB),
	get_form_value(Input, language, Lang),
	current_output(OldOutput),
	open('/tmp/temp_qfile', write, S),
	set_output(S),
	write(AtomQuery), 
	(Lang=on-> write('.');true),
	nl,
	set_output(OldOutput),
	close(S),
	open('/tmp/temp_qfile', read, Stream),
	error_protect(read_term(Stream, Query,
                      [variables(Vs), variable_names(Ns)])),
	delete_file('/tmp/temp_qfile'),		       
	(Lang=on ->
	 error_protect(sql_query(DB, Query, R)),
	 R=..[_|[TermList]],
	 TermList=[Tuple|_],
	 length(Tuple, Arity),
	 process_output([], TermList, Arity, Reply, Response)
	 ;
	 validate(Query, DB, Res),
	 (Res=ok ->
	  R=..[r|Vs],
	  functor(R,r,Arity),
	  error_protect(itf_findall(DB, R, Query, L)),
	  order_var_names(Ns, Vs, NsFinal),
	  process_output(NsFinal, L, Arity, Reply, Response)
	 ;
	  text_option(unsuccessful_query,_, Unsuccesful),	    
	  html_option(admin_short_message, [Unsuccesful], Reply),
	  Response=[])
	),
	update_stats(_, c).
process_query(_, Reply, []):-
	text_option(unsuccessful_query,_, Unsuccesful),	    
	html_option(admin_short_message, [Unsuccesful], Reply).

:- pred validate(+Query, +DB, -Res)

   # "If in prolog syntax, checks the predicates involved in the query 
      @var{Query} are all @pred{authorized/1}. ".

validate(Query, DB, Res):-
	functor(Query, F, _A),
	(F==',';F==';'), !,
	arg(1, Query, SubQ1),
	arg(2, Query, SubQ2),
	validate(SubQ1, DB, Res1),
	validate(SubQ2, DB, Res2),
	(Res1=ok, Res2=ok ->
	 Res=ok
	;
	 Res=ko).
validate(Query, DB, Res):-
	functor(Query, F, A),
	(authorized(F/A, DB)->
	 Res=ok
	;
	 Res=ko).	

authorized(F/A,_):-
	dbtable(F,A,_,_,_),!.
authorized(F/A, DB):-
	\+(DB=local),
	atom_concat([F,'_',DB], F_DB),
	dbtable(F_DB,A,_,_,_),!.
authorized('+'/2,_):-!.
authorized('-'/2,_):-!.
authorized('*'/2,_):-!.
authorized('/'/2,_):-!.
authorized('\+'/2,_):-!.
authorized('='/2,_):-!.
authorized('=='/2,_):-!.
authorized('>'/2,_):-!.
authorized('<'/2,_):-!.
authorized('>='/2,_):-!.
authorized('=<'/2,_):-!.
authorized('is'/2,_):-!.
authorized('=:='/2,_):-!.
authorized('=/='/2,_):-!.

order_var_names(_,[],[]).
order_var_names(NamesEqVar, [GuideVar|Vs], [Name|RNs]):-
	member_esp((Name=GuideVar), NamesEqVar),
	order_var_names(NamesEqVar,Vs, RNs).

member_esp((X=Z), [(X=Y)|_]):- Z==Y, !.
member_esp(X, [_|R]):- member_esp(X,R).

:- pred process_output(+Vars, +Rows, +Arity, -Reply, -Page)

   # "Formats the output of a given query in Prolog/SQL syntax. ".

process_output(_,[],_, Reply, []):- !,
	text_option(no_found_reply, _, Mess),
	html_option(admin_short_message, [Mess], Reply).
process_output(Vs, Rs, Arity, Reply, Page):-
	name_list_to_table_heading(Vs, Heading),
	data_to_table_row(Heading, Row1),
	get_general_rows(Rs, RRows),
	Rows=[Row1|RRows],
	number_codes(Arity, A),
	atom_codes(AtomArity, A),
	html_option(general_table, [AtomArity, Rows], Page),
	html_option(admin_short_message, ['Complex query resulted in:'], Reply).

get_general_rows([], []). 
get_general_rows([I1|Is], [R1|Rs]):- 
	(I1=[_|_] ->
	 Args=I1
	;
	 I1=..[_|Args]
	),
	data_to_row_elements(Args, RowData),
	data_to_table_row(RowData, R1),
	get_general_rows(Is, Rs).


:- pred tables_locations(+Locations, -Menu)

   # "Displays a menu where the user can choose the location a table of a 
      database belongs to. ".

tables_locations([],[]).
tables_locations([local|Locations], [O1|Opts]):-
	!,O1 = option([value=local], local),
	tables_locations(Locations, Opts).
tables_locations([L1|Locations], [O1|Opts]):-
	sql_persistent_location(L1,db(_,U, _, _:_)),
	O1 = option([value=L1], U),
	tables_locations(Locations, Opts).

:- pred display_tables_info(+Tables, -TablesInfo)

   # "Displays a table with information about the tables contained in the 
      system, their number of arguments and their field types. ".

display_tables_info([], []).
display_tables_info([table_info(Table_Name, Arity, TFD, Site)|TInfos], 
	            [Table_Info_Html|Ts]):-
	table_info(Table_Name, Arity, TFD, Site, Table_Info_Html),
	display_tables_info(TInfos, Ts).

table_info(TN, A, TFD, Site,THTML):-
	number_codes(A, NA),
	atom_codes(CA, NA),
	(Site = local ->
	 TDN = TN,
	 U = Site
	;
        atom_concat('_',Site, TNAux),
	 atom_concat(TDN, TNAux, TN),
    	 sql_persistent_location(Site, db(_, U, _, _:_))
	),
	atom_concat([TDN, '/', CA, ', ', 'laying on view: ', U], PredName),
	html_option(table_heading_elem, [PredName], NameElem),
	TFD=..[_|FieldsTypes],
	format_types(FieldsTypes, TFields),
	html_option(table_element, [TFields], TFs),
	html_option(data_row, [NameElem, TFs], THTML).

format_types([], []).
format_types([Fd|FieldsTypes], [Tf |TFields]):-
	Fd=..[Functor|Args],
	concat_args(Args, CArgs),
	atom_concat([Functor, '(', CArgs, ')'], AFd),
	html_option(field_desc, [AFd], Tf),
	format_types(FieldsTypes, TFields).

concat_args([A], CA):-
	get_atom_with_suffix(A,'',CA).
concat_args([A|As], CAs):-
	get_atom_with_suffix(A, ',', CA),
	concat_args(As, CAsAux),
	atom_concat(CA, CAsAux, CAs).

get_atom_with_suffix(Code, Suffix, Atom):-
	atom(Code), !,
	atom_concat([Code, Suffix], Atom).
get_atom_with_suffix(Number, Suffix, Atom):-
	number_codes(Number, CNumber),
	atom_codes(ANumber, CNumber),
	atom_concat(ANumber, Suffix, Atom).

:- pred updatetemplates(+ClientType, -Response) :: passwd * output_page

   # "This service is used to reload the HTML templates, which control
      the settings, or any of the files that define the types used,
      after having edited any of them. This way, the system
      administrator can customize the system on-line, with no need to
      stop it.  Only authorised clients can use this service.".

updatetemplates(Passwd, Response) :-
       check_passwd(Passwd),
       db_templates(Dir),
       read_option_html(_, Dir),
       create_ini_page,
       include_parametric_files,
       text_option(templates_title,_, Title),
       text_option(templates_reloaded,_, Message),
       db_url(URL),
       message_page(Title, Message, URL, Response).

:- pred consultdboptions(+Input, +ClientType, -Response) 
   :: dic * passwd * output_page

   # "Implements the online system messages edit service. Provides an
      interface to edit on line every system message and, in general,
      any text which appears in the System's user interface. For
      example, the text of any link, the heading or title of any page
      or the message given when an element has been successfully
      inserted.  Only authorised clients can use this service.".

consultdboptions(Input, Passwd, Response) :-
        check_passwd(Passwd),
        get_form_value(Input, selected, yes), !,
        update_changed_values(Input),
        create_ini_page_if_needed,
        text_option(options_title,_, Title),
        text_option(options_changed,_, Message),
	db_url(URL),
        message_page(Title, Message, URL, Response).
consultdboptions(_Input,_Passwd, Response) :-
	findall(Option, option_entry(Option), FOps),
	html_option(options_body,[FOps], AllOps),
	text_option(options_title,_, Title),
	text_option(option_message,_, OMText),
	html_option(general_add,[OMText], Message),
	html_option(admin_heading, [Title], Head),
	html_option(admin_content, [Change_option], Content),
	html_option(admin_footer, [], Footer),        
	html_option(ok_button,[], Ok_Button),
	html_option(reset_button, [], Reset_Button),
	html_option(form_buttons, [Ok_Button, Reset_Button], Buttons),
	platform_cgi('admin/db_generic_client', CGI),
	Change_option = [
			    Message,
			    start_form(CGI),
			    AllOps,
			    input(hidden, [name=selected, value=yes]),
			    input(hidden, [name=service, 
			                   value=modify_options_service]),
                            Buttons,
			    end_form
                        ],
	html_option(page_body, [[Head, Content, Footer]], Body),
	full_reply_page(Title, Body, Response).

update_changed_values(Input) :-
       text_option(Name, Desc, Val),
       member(Name=NewVal, Input),
       text_update(Val, NewVal, Name, Desc),
       check_ini_page_change(Name),
       fail.
update_changed_values(_).

text_update(Val, NewVal,_,_) :-
	Val == NewVal, !.
text_update(_, NewVal, Name, Desc) :-
       persdbrt:retract_fact(text_option(Name,_,_)),
       persdbrt:assertz_fact(text_option(Name, Desc, NewVal)).

:- pred create_destroy_table(+Input, +ClientType, -Response) 
   :: dic * passwd * output_page

   # "Implements the Create/Destroy database table service. 
      Provides an interface where the user can create a new table or
      destroy an already existing one, or both at the same
      time. Tables created this way have arity zero and need to be
      subsequently edited through the service @pred{edit_table/3}.
      Only authorised clients can use this service.".

create_destroy_table(Input,Passwd, Response) :-
       check_passwd(Passwd),		
       operate_on_tables(Input),
       findall(TI, dbtable(TI, _, _, _, _), TIList),
       table_options(TIList, Options),
       text_option(table_creation_title,_, Title),
       html_option(general_heading, [Title], Mess),
       text_option(create_table_message,_, Message),
       text_option(destroy_table_message,_, Message2),
       html_option(table_element, [Message],Create_Mess),
       html_option(table_element, [Message2],Destroy_Mess),
       html_option(table_element, [input(text, [name=new_table_name, size=40])], CreateInput),
       html_option(table_element, [select([name=list],
        [option([value=not_destroy], 'Keep current tables')|Options])],Destroy_List),
       data_to_table_row([Create_Mess, CreateInput], CreateRow),
       data_to_table_row([Destroy_Mess, Destroy_List], DestroyRow),
       html_option(general_table, ['2', [CreateRow, DestroyRow]], Creation_table),
       html_option(admin_heading, [Title], Head),
       html_option(admin_content, [Add_table], Content),
       html_option(admin_footer, [], Footer),        
       html_option(ok_button, [], Ok_Button),
       html_option(reset_button, [], Reset_Button),
       html_option(form_buttons, [Ok_Button, Reset_Button], Buttons),
       platform_cgi('admin/db_generic_client', CGI),
       Add_table = [
		       Mess,
		       start_form(CGI),
		       Creation_table,			 
		       input(hidden, [name=selected, value=yes]),
		       input(hidden, [name=service, 
		                      value=create_table_service]),
		       Buttons,
		       end_form
		   ],
	html_option(page_body, [[Head, Content, Footer]], Body),
	full_reply_page(Title, Body, Response).

:- pred operate_on_tables(+Input)

   # "From the user interface it gets which table has to be added or destroyed 
      or both in the system and does so. ".

operate_on_tables(Input):-
	get_form_value(Input, selected, yes), !,
	get_form_value(Input, new_table_name, New_Table_Name),
	get_form_value(Input, list, Destroy_Table), 
	add_new_table(New_Table_Name),
	destroy_table(Destroy_Table).
operate_on_tables(_).

add_new_table(Table_Name):-
	form_empty_value(Table_Name), !.
add_new_table(Table_Name):-
	date(Date),
	persdbrt:assertz_fact(dbtables_desc(Table_Name, local, Date)),
	data(Table_Name/0),
	make_persistent(Table_Name/0, db),
	data_facts:assertz_fact(dbtable(Table_Name, 0, Table_Name, local, Date)),
	initialize_stat(Table_Name).

destroy_table(not_destroy):- !.
destroy_table(TI):-
	dbtable(TI, Arity, FieldsDesc, local, Date),!,
	get_free_vars(1, Arity, VarList),
	Pred=..[TI|VarList],
	pretract_all(Pred),
	persistent_dir(db, Dir),
	number_codes(Arity, NA),
	atom_codes(CA, NA),
	atom_concat([Dir, 'dbserver:', TI, '-', CA, '.pl'], File),
	atom_concat([Dir, 'dbserver:', TI, '-', CA, '_ops.pl'], File_ops),
	atom_concat([Dir, 'dbserver:', TI, '-', CA, '_bak.pl'], File_bak),
	delete_file1(File),
	delete_file1(File_ops),
	delete_file1(File_bak),
	persdbrt:retract_fact(dbtables_desc(FieldsDesc, local, Date)),
	data_facts:retract_fact(dbtable(TI, Arity, FieldsDesc, local, Date)),
	reset_stat(TI).	
destroy_table(TI):-
	dbtable(TI, Arity, FieldsDesc, SiteId, Date),
	persdbrt:retract_fact(dbtables_desc(FieldsDesc, SiteId, Date)),
	data_facts:retract_fact(dbtable(TI, Arity, FieldsDesc, SiteId, Date)),
	reset_stat(TI),!,
	(call(dbtable(_,_,_,SiteId,_)) ->
	 true
	;
	 unlink_DB(SiteId)).
destroy_table(_).

pretract_all(P):-
	persdbrt:retract_fact(P),
	fail.
pretract_all(_).

:- pred select_table_to_edit(+Passwd, -Response) :: passwd * output_page

   # "Implements with @pred{edit_table/3} the edit table type service. 
      Similar to @pred{get_relationsdb/2}, provides a menu to choose the
      table whose field types the user wants to edit.  Only authorised
      clients can use this service.".

select_table_to_edit(Passwd, Reply):-
	check_passwd(Passwd),
	text_option(edit_table_title,_, Title),
	text_option(edit_table_heading,_,Heading),
	html_option(admin_heading, [Heading], Head),
	html_option(admin_content, [Choose_table], Content),
	text_option(choose_table_edit_mess, _,CTText),
	html_option(general_choose,[CTText],Message),
	html_option(ok_button,[],Ok_Button),
	html_option(admin_footer, [], Footer),
	html_option(page_body, [[Head, Content, Footer]], Body),
	findall(TI, dbtable(TI, _, _, local,_), TIList),
	table_options(TIList, Options),
	platform_cgi('admin/db_generic_client', CGI),
	Choose_table = [
			   Message,
			   start_form(CGI),
			   select([name=table_name], Options),
			   input(hidden, [name=initial, value=yes]),
			   input(hidden, [name=service, 
			                  value=table_to_edit_service]),
                        \\, \\,
                        Ok_Button,
                        end_form
                      ],
        full_reply_page(Title, Body, Reply).

:- pred edit_table(+Input, +Passwd, -Response) :: dic * passwd * output_page 

   # "Once the user has selected through @pred{select_table_to_edit/2}
      the table to be edited, provides an interface which allows the
      user to edit any type of its fields or add a new one with a
      given type.  Only authorised clients can use this service.".

edit_table(Input, Passwd, Response):-
	check_passwd(Passwd),
	get_form_value(Input, table_name, Table),
	set_edit_answer(Input, Table, Table_out, Message, Edition_table),
	html_option(admin_footer, [], Footer),
	text_option(edit_table_title,_, Title),
	text_option(edit_table_heading,_,Heading),
	html_option(page_body, [[Head, Content, Footer]], Body),
	html_option(admin_heading, [Heading], Head),
	html_option(admin_content, [Edit_table], Content),        
	html_option(ok_button,[],Ok_Button),
	html_option(reset_button,[],Reset_Button),
	html_option(form_buttons,[Ok_Button, Reset_Button], Buttons),
	text_option(add_field_message,_, AFText),
	html_option(general_add, [AFText], Add_field_message),
	html_option(table_element, 
	 [input(text, [name=add_field_desc, size=25])], NewFieldDesc),
	findall(option([value=Option], Option), 
	 (option_table_size(Option,_), Option\==none), Options),
	html_option(table_element, 
	 [select([name=add_field_type],[none|Options])], NewFieldType),
	data_to_table_row([NewFieldDesc,NewFieldType], Add_field_row),  
	html_option(general_table, ['2', Add_field_row], Added_row_table),
	platform_cgi('admin/db_generic_client', CGI),
	Edit_table=[
		       Message,
		       start_form(CGI),
		       Edition_table,
		       \\, \\,
		       Add_field_message,
		       Added_row_table,
		       input(hidden, [name=table_name, value=Table_out]),
		       input(hidden, [name=service, 
			               value=table_to_edit_service]),	
			Buttons,
			end_form
		    ],
        full_reply_page(Title, Body, Response).	

set_edit_answer(Input, TI, TI_out, Message, Edition_table):-
	set_edit_mode(Input, TI, TI_out),
	dbtable(TI_out, Arity, TableDescFields, local,_),
	TableDescFields=..[_|Types],
	get_field_types(1, Types, Rows),
	(Arity = 0 ->
	 Message = []
	;
	 text_option(edit_table_message, _, ETText),
	 html_option(general_choose,[ETText],Message)
	),
	html_option(general_table, ['2', Rows], Edition_table).

set_edit_mode(Input, TI, TI):-
	get_form_value(Input, initial, yes), !.
set_edit_mode(Input, TI, TI_out):-
	update_table(Input, TI, TI_out).

update_table(Input, Table, TI_out):-
	get_form_value(Input, table_name, Table),
	get_form_value(Input, add_field_desc, New_desc),
	get_form_value(Input, add_field_type, New_type),
	dbtable(Table, Arity, _, local, _),
	get_n_preds_form_vals(Input, edit_line, 2, Arity, LLVals),
	(form_empty_value(New_desc) ->
	 compose_table_types(LLVals, TableTypes)
	;
	 append(LLVals, [[New_desc, New_type]], New_pred_desc),
	 compose_table_types(New_pred_desc, TableTypes)
	),
	NewFieldsDesc=..[Table|TableTypes],
	edit_dbtables(Table, NewFieldsDesc, TI_out).
	
compose_table_types([], []).
compose_table_types([[_Desc,none]|LLVs], TTs):-
	compose_table_types(LLVs, TTs).
compose_table_types([[Desc,TypeOption]|LLVs], [TypeTerm|TTs]):-
	option_table_size(TypeOption, TypeTerm),
	(form_empty_value(Desc) ->
	 arg(1, TypeTerm, anonymous)
	;
	 arg(1, TypeTerm, Desc)
	),
	compose_table_types(LLVs, TTs).
	
:- pred edit_dbtables(+TI, +NewFieldsDesc, -TI)

   # "Actually updates a table after having changed it description. The old 
      data is inconsistent, therefore eliminated. ".

edit_dbtables(TI, NewFieldsDesc, TI):-
	dbtable(TI, _Arity, FieldsDesc, local,_),
	NewFieldsDesc = FieldsDesc, !.
edit_dbtables(TI, NewFieldsDesc, NTI):-
	dbtable(TI, Arity, FieldsDesc, local, _),
	persdbrt:retract_fact(dbtables_desc(FieldsDesc, local, _)),
	date(Date),
	persdbrt:assertz_fact(dbtables_desc(NewFieldsDesc, local, Date)),
	get_free_vars(1, Arity, VarList),
	Pred=..[TI|VarList],
	pretract_all(Pred),
	functor(NewFieldsDesc, _F, NewArity),
	data_facts:assertz_fact(dbtable(TI,NewArity,NewFieldsDesc,local,Date)),
	update_dyn_pred(TI, Arity, NewArity),
	NTI=TI,
	persistent_dir(db, Dir),
	number_codes(Arity, NA),
	atom_codes(CA, NA),
	atom_concat([Dir, 'dbserver:', TI, '-', CA, '.pl'], File),
	atom_concat([Dir, 'dbserver:', TI, '-', CA, '_ops.pl'], File_ops),
	atom_concat([Dir, 'dbserver:', TI, '-', CA, '_bak.pl'], File_bak),
	delete_file1(File),
	delete_file1(File_ops),
	delete_file1(File_bak), 
	data_facts:retract_fact(dbtable(TI, Arity, FieldsDesc, local, _)),
	reset_stat(TI),
	initialize_stat(TI),!.
edit_dbtables(_, _, _).

delete_file1(File):-
	(file_exists(File) ->
	 delete_file(File)
	;
	 true).

get_free_vars(N, A, []):- N > A, !.
get_free_vars(N, A, [V|Vs]):-
	var(V),
	N1 is N + 1,
	get_free_vars(N1, A, Vs).

update_dyn_pred(P, A, NA):-
	A \== NA, !,
	data(P/NA),
	make_persistent(P/NA, db).
update_dyn_pred(_, _, _).

get_field_types(_, [], []).
get_field_types(Index, [Type|Ts], [Edit_line|Es]):-
	type_info(Type, Format, Desc),
	process_type(Format, Option),
	functor(Type, F, _N),
	(Option='' ->
	 FinalOption=F
	;
	 atom_concat([F, '_', Option], FinalOption)
	),
	make_edit_line(Index, Desc, FinalOption, Edit_line),	
	NewIndex is Index + 1,
	get_field_types(NewIndex, Ts, Es).	

:- pred make_edit_line(+Index, +Desc, +Opt, -Edit_line)

   # "Provides a new part of the interface where the user can write the name
      of the new field and choose its type. ".

make_edit_line(Index, Desc, Opt, Edit_line):-
	compose_table_row_field_names(edit_line,  Index, 2, [Field1, Field2]),
	findall(option([value=Option], Option),
	        option_table_size(Option, _), OptionList),
	select(option([value=Opt], Opt), OptionList, OptionRest),
	NewOptList=[option([value=Opt], Opt)|OptionRest],
	html_option(table_element, 
	 [input(text, [name=Field1, size=25, value=Desc])], RowField1),
	html_option(table_element, 
	 [select([name=Field2], NewOptList)], RowField2),
	data_to_table_row([RowField1, RowField2], Edit_line).

:- pred process_type(+Format, -Option)

   # "Given an HTML format @var{Format}, obtains a text description for 
      the size of the input field that correlates with it. ".

% process_type(Format, Option):-
% 	html_format(Format,_,_,unidimensional,_), !,
% 	arg(2, Format, Length),
% 	get_option_size(Length, Option).
% process_type(Format, Option):-
% 	html_format(Format,_,_,bidimensional,_), !,
% 	arg(2, Format, Rows),
% 	arg(3, Format, Cols),	
% 	get_option_bi_size(Rows, Cols, Option).
% process_type(Format, ''):-
% 	html_format(Format,_,_,plain,_).

process_type(Format, Option):-
	html_format_input_dimensions(Format, uni(Length)), !,
	get_option_size(Length, Option).
process_type(Format, Option):-
	html_format_input_dimensions(Format, bi(Rows, Cols)), !,
	get_option_bi_size(Rows, Cols, Option).
process_type(_Format, '').

get_option_size(L, 'small (1-20)'):- L > 0, L < 21, !.
get_option_size(L, 'medium (21-40)'):- L > 20, L < 41, !.
get_option_size(L, 'large (41-60)'):- L > 40, L < 61, !.
get_option_size(_, none).

get_option_bi_size(R, C, 'small (Rows*Cols <= 200)'):- R*C < 201, !.
get_option_bi_size(R, C, 'medium (Rows*Cols <= 400)'):- R*C < 401, !.
get_option_bi_size(R, C, 'large (Rows*Cols <= 1200)'):- R*C < 1201, !.
get_option_bi_size(_, _, none).

:- data ini_page_changed/0.

check_ini_page_change(Name) :-
        (ini_page_text_option(Name) ->
	  data_facts:asserta_fact(ini_page_changed)
        ; 
	  true).

:- pred create_ini_page_if_needed 

   # "If anything in the graphic display of the system's front page has changed
      the system finds it out automatically and generates it over again. ".

create_ini_page_if_needed :-
        ini_page_changed,
        data_facts:retractall_fact(ini_page_changed),
        create_ini_page.
create_ini_page_if_needed.

option_entry(Entry) :-
        text_option(Name, Desc, Val),
        html_option(option_entry,[Name,Desc,Val],Entry).

:- pred update_stats(+TI, +Action)

   # "Updates the system statistics for every action and database table. ".

update_stats(TI, s):- !,
	itf_retract(table_stats(TI, S, I, D, M,  _)),
	inc_field_info(S, New_S, Date),
	itf_assert(table_stats(TI, New_S, I, D, M, Date)).
update_stats(TI, i):- !,
	itf_retract(table_stats(TI, S, I, D, M, _)),
	inc_field_info(I, New_I, Date),
	itf_assert(table_stats(TI, S, New_I, D, M, Date)).
update_stats(TI, d):- !, 
	itf_retract(table_stats(TI, S, I, D, M, _)),
	inc_field_info(D, New_D, Date),
	itf_assert(table_stats(TI, S, I, New_D, M, Date)).
update_stats(TI, m):- !,
	itf_retract(table_stats(TI, S, I, D, M, _)),
	inc_field_info(M, New_M, Date),
	itf_assert(table_stats(TI, S, I, D, New_M, Date)).
update_stats(_, c):- !,
	itf_retract(table_stats('$Complex queries', S, I, D, M, _)),
	inc_field_info(S, New_S, Date),
	itf_assert(table_stats('$Complex queries', New_S, I, D, M, Date)).
update_stats(_, _).

inc_field_info(N, N1, Date):-
	N1 is N + 1,
	date(Date).

:- pred status(Response) => output_page

   # "Implements the system status service. Returns information about the 
      system itself and the tables it holds.".

status(Response) :-
       text_option(status_title,_, Title),
       text_option(status_message,_, SMText),	
       launch_date(LDate),
       atom_concat(SMText, LDate, SSText),
       html_option(general_add,[SSText], Message),
       html_option(admin_heading, [Title], Head),
       html_option(page_body, [[Head, Content, []]], Body),
       html_option(admin_content, [Status], Content),
       findall(t(T,S,I,D,M,CD,L), (table_stats(T,S,I,D,M,L), 
       dbtable(T,_,_,_,CD)), List),
       table_stats('$Complex queries', Sc, Ic, Dc, Mc, Lc),
       text_option(c_q_access,_,CQH),
       StatsList=[t(CQH, Sc, Ic, Dc, Mc, Lc, Lc)|List],
       text_option(search_access,_,Search),
       text_option(insert_access,_,Insert),
       text_option(delete_access,_,Delete),
       text_option(modify_access,_,Modify),	
       text_option(created_on,_,Created),
       text_option(last_accessed,_,Last),	
       text_option(table_prompt,_,TName),	
       text_option(total_access,_,TotalPrompt),		
       text_option(global_stats_heading,_,GSH),
       text_option(det_stats_heading,_,DSH),
       html_option(admin_heading, [GSH], GSHead),
       html_option(admin_heading, [DSH], DSHead),
       name_list_to_table_heading([TName,Search,Insert,Delete,Modify,Created,Last], THeading),
       get_general_rows(StatsList, RRows),
       data_to_table_row(THeading, R1),
       Rows=[R1|RRows],
       html_option(general_table,['8',Rows], TStats),	
       get_total_parts(StatsList, 0, 0, 0, 0, St, It, Dt, Mt),
%	get_graphic(St, It, Dt, Mt, Graphic),
       TotalAccesses is St + It + Dt + Mt,
       name_list_to_table_heading([Search, Insert, Delete, Modify, TotalPrompt], TotalHeading),
       data_to_table_row(TotalHeading, Tr1),
       get_general_rows([[St, It, Dt, Mt, TotalAccesses]], Trow),
       Trows=[Tr1|Trow],
       html_option(general_table,['5',Trows], GTStats),
       Status=[
		  GSHead,
		  Message,
%		  Graphic,
		  GTStats,
		  DSHead,
		  TStats
              ],
       full_reply_page(Title, Body, Response).

get_total_parts([], S, I, D, M, S, I, D, M).
get_total_parts([t(_, S, I, D, M,_,_)|RTs ], Sx, Ix, Dx, Mx, Sf, If, Df, Mf):-
	Saux is Sx + S,
	Iaux is Ix + I,
	Daux is Dx + D,
	Maux is Mx + M,
	get_total_parts(RTs, Saux, Iaux, Daux, Maux, Sf, If, Df, Mf).

:- pred get_graphic(+St, +It, +Dt, +Mt, -Graph)

   # "Obtains a graphic representation of the number of actions (searches, 
      insertions, deletions and modifications). ".

get_graphic(St, It, Dt, Mt, Graph):-
	current_output(OldOutput),
	open('/tmp/data', write, S),
	set_output(S),
	write_data_to_file(1,4,[St, It, Dt, Mt]),
	set_output(OldOutput),
	close(S),
	current_executable(Exe),
	absolute_file_name(Exe,'','','.',_,_,Dir),
	atom_concat(['gnuplot ',Dir,'/graph '], StatScript),
	system(StatScript),
	db_dir(DBDir),
	atom_concat(['mv -f /tmp/graph.gif ', DBDir, '/stats'], MVCommand),
	system(MVCommand),
	html_option(stats, ['stats/graph.gif'], Graph),!.
get_graphic(_, _, _, _, []).

write_data_to_file(M,N,[D|Ds]):-
	M<N+1, !,
	write(M), write(' '), write(D), nl,
	M1 is M+1,
	write_data_to_file(M1, N, Ds).
write_data_to_file(_, _, []).

:- comment(die/1,"Once the database has been updated in order to keep
   all the information in a consistant state, it halts the database
   server.  Only authorised clients can use this service.").

:- pred die(ClientType) : passwd

   # "Used to kill the server (active module) after updating the database.".

die(Passwd) :-
	check_passwd(Passwd), 
	initialize_db,
	update_db,
	get_addr_dir(DirAddr,_),
	get_lock_file(DirAddr, Lock_File),
	delete_file(Lock_File),
	halt.

update_db:-
	dbtable(T,A,_, _, _),
	make_persistent(T/A, db),
	fail.
update_db.

:- pred dying(+ClientType, -Response) :: passwd * output_page

   # "Implements with @pred{die} the stop the system service. 
      Before having halted the database server through @pred{die/1},
      this predicate is used to produce a message, pointing out that
      the system is about to be halted.  Only authorised clients can
      use this service.".

dying(Passwd, Response) :-
	check_passwd(Passwd), 	
	text_option(stop_title,_, Title),
	text_option(stop_message,_, Message),
	db_url(URL),
	message_page(Title, Message, URL, Response).

% % Moved to html_generation.pl
% message_page(Title, Message, BaseURL, Response) :-
% 	html_option(page_body,_, [[Head, Content]], Body),
% 	html_option(admin_heading,_, [Title], Head),
% 	html_option(admin_short_message,_,[Message], Mess),
% 	html_option(admin_content,_, [Mess], Content),
% 	compose_page(Title, Body, BaseURL, Response).

check_passwd(Passwd) :-
	nonvar(Passwd), password(Passwd), !.
check_passwd(Bad) :-
	bad_passwd(Bad).

bad_passwd(Bad) :-
       inform_user(['WARNING! Bad passwd sent: ',Bad,
                     ' ; maybe somebody is attacking the system']),
       halt.

get_addr_dir(DirAddr, Exe) :-
	current_executable(Exe),
	absolute_file_name(Exe,'','','.',_,_,Dir),
	atom_concat(Dir,'/.addr',FAddr),
	open_input(FAddr, SFAddr),
	read(DirAddr),
	close_input(SFAddr).

get_lock_file(DirAddr, LockFile) :-
	atom_concat(DirAddr,'/.lock', LockFile).

:- initialization(startupdb).

:- pred startupdb # "Reads the configuration files of the database. If
   the source file is found, it creates the maintenance and common client 
   front pages and dies.".

startupdb:-
	initialization(operators),
	umask(_, 0o007),
	get_addr_dir(DirAddr, E),
	get_lock_file(DirAddr, Lock_file),
	get_pid(PID),
	open_output(Lock_file, SS),
	write(PID),
	close_output(SS),
	atom_concat(E, '_ini.pl', File),
	open_input(File, IO), !,
 	repeat,
	  read(T),
	  process(T),
	T == end_of_file, !,
	close_input(IO),
	make_desc_persistent,
	include_parametric_files,
	db_templates(Dir),
	read_option_html(_,Dir),
	(exists_pl(E) ->
	 init_persdb,
	 read_db_tables,
	 create_ini_page,
	 password(Passwd),
	 die(Passwd)
       ; 
	 retractall(dbtable(_,_,_,_,_)),
	 init_persdb,
	 include_sql_locations,
	 read_db_tables
       ),
       date(Date),
       data_facts:asserta_fact(launch_date(Date)),
       initialize_stat('$Complex queries').

exists_pl(E) :-
	atom_concat(E,'.pl',File),
	file_exists(File).
% Kludge for windows
exists_pl(E) :-
	atom_concat(E1,'.cpx', E),
	atom_concat(E1, '.pl', File),
	file_exists(File).

process(end_of_file) :- !.
process(db_os(OS)) :- !,
	data_facts:asserta_fact(db_os(OS)).
process(db_dir(Dir)) :- !,
        add_final_slash(Dir, DIR),
	data_facts:asserta_fact(db_dir(DIR)).
process(db_types(Dir)) :- !,
        add_final_slash(Dir, DIR),
	data_facts:asserta_fact(db_types(DIR)).
process(db_templates(Dir)) :- !,
        add_final_slash(Dir, DIR),
	data_facts:asserta_fact(db_templates(DIR)).
process(db_url(Dir)) :- !,
        add_final_slash(Dir, DIR),
	data_facts:asserta_fact(db_url(DIR)).
process(db_description(Dir)) :- !,
        add_final_slash(Dir, DIR),
	data_facts:asserta_fact(persistent_dir(db_description, DIR)).
process(db_data(Dir)) :- !,
        add_final_slash(Dir, DIR),
	data_facts:asserta_fact(persistent_dir(db, DIR)).
process(Bad) :-
	inform_user(['Bad data in init file: ',Bad]).

:- data text_option/3.
:- data search/3.
:- data sql_location/2.
:- data table_stats/6.
:- data dbtables_desc/3.

make_desc_persistent :-
	make_persistent(text_option/3, db_description),
	make_persistent(search/3, db_description),
	make_persistent(sql_location/2, db_description),
	make_persistent(table_stats/6, db_description),
	make_persistent(dbtables_desc/3, db_description).

add_final_slash(Dir, DIR) :-
        atom_concat(_,'/',Dir) -> DIR = Dir ; atom_concat(Dir, '/', DIR).

include_sql_locations:-
	sql_location(DBId, DBHandle),
	data_facts:assertz_fact(sql_persistent_location(DBId, DBHandle)),
	fail.
include_sql_locations.

read_db_tables :-
	dbtables_desc(Table, SiteId, Date),
	functor(Table, F, A),
	(SiteId = local ->
	 data_facts:assertz_fact(dbtable(F, A, Table, local, Date)),
	 data(F/A),
	 TName=F,
	 make_persistent(F/A, db) 
	;
	 atom_concat([F,'_',SiteId], TName),
	 data_facts:assertz_fact(dbtable(TName, A, Table, SiteId, Date)),
	 get_sql_args(Table, PrologDef, SQLDef),	 
	 make_sql_persistent(PrologDef, SQLDef, SiteId)
	),
	initialize_stat(TName),
	fail.
read_db_tables.

initialize_stat(TName):-
	call(table_stats(TName,_,_,_,_,_)), !.
initialize_stat(TName):-
	date(Date),
	itf_assert(table_stats(TName,0,0,0,0,Date)).

reset_stat(TName):-
	itf_retract(table_stats(TName,_,_,_,_,_)).

reset_stats(DbId):-
	dbtable(TI,_,_,DbId,_),
	reset_stat(TI),
	fail.
reset_stats(_).

include_parametric_files:-
%	include_parametric_file('html_format.pl', html_format(_,_,_,_,_)),
% Moved to html_generation.pl
	db_types(Dir),
	load_html_format(html_format, Dir),
	include_parametric_file('html2sql.pl', html2sql(_,_)),
	include_parametric_file('sql2html.pl', sql2html(_,_,_)),
	include_parametric_file('option_table_size.pl',option_table_size(_,_)),
	include_parametric_file('dbconn_desc.pl',dbconn_desc(_,_,_,_,_,_)),
	include_parametric_file('search_expire.pl',search_expire(_)).

include_parametric_file(File, Pattern):-
	db_types(Dir),
	atom_concat(Dir, File, CFile),
	(data_facts:retractall_fact(Pattern);true),
	open_input(CFile, IO), !,
	read(Fact),
	process_file(Fact),
	close_input(IO).

include_parametric_file(_,_).

process_file(end_of_file):- !.
process_file(Fact):-
	data_facts:assertz_fact(Fact),
	read(NewFact),
	process_file(NewFact).

%%%%%%%%%%%%%%% System utility predicates %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% date(Date):-
%         current_executable(Exe),
%         absolute_file_name(Exe,'','','.',_,_,Dir),
%         atom_concat(Dir,'/date.sh', DateScript),
% 	popen(DateScript, read, S),
% 	read(S, Date).

date(Date) :-
	datime(X),
	X=..[_|DTList],
	get_atoms(DTList, [Year, Month, Day, Hour, Min, Sec]),
	atom_concat([Year, '/', Month, '/', Day, '-', 
	            Hour, ':', Min, ':', Sec], Date).

get_atoms([], []).
get_atoms([N|Ns], [A|As]) :-
	number_codes(N,NList), 
	atom_codes(A, NList),
	get_atoms(Ns, As).

extract_preds([], []).
extract_preds([pred(_N,Pred)|Ps], [Pred|EPs]) :-
	extract_preds(Ps, EPs).
	
%%%%%%%%%%%%%%% Special predicates for persdb_sql handling %%%%%%%%%%%%%%%%%%%%
get_sql_args(Table, PrologDef, SQLDef):-
	Table=..[TName|TFields],
	get_fields_ids_types(TFields, IDs, Types),
	SQLDef=..[TName|IDs],
	PrologDef=..[TName|Types].

get_fields_ids_types([], [], []).
get_fields_ids_types([T1|Types], [Id1|IDs], [TT1|TTs]):-
	type_info(T1, _, Id1),
	html2sql(T1, TT1),
	get_fields_ids_types(Types, IDs, TTs).

%%%%%%%%%%%%%%%%%%%%%%%%%% Interface predicates %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
itf_assert(Tuple):-
	dbassertz_fact(Tuple), !.
itf_assert(Tuple):-
	persdbrt:assertz_fact(Tuple), !.
itf_assert(Tuple):-
	data_facts:assertz_fact(Tuple).

itf_retract(Tuple):-
	dbretract_fact(Tuple), !.
itf_retract(Tuple):-
	persdbrt:retract_fact(Tuple), !.
itf_retract(Tuple):-
	data_facts:retract_fact(Tuple), !.
itf_retract(_).

itf_retractall([]).
itf_retractall([P|Ps]):-
	itf_retract(P),
	itf_retractall(Ps).

itf_retractall_fail(Fact):-
	itf_retract(Fact),
	fail.
itf_retractall_fail(_).

itf_findall(local, Recipient, Template, ResultList):-
	!, findall(Recipient, Template, ResultList).
itf_findall(DBId, Recipient, Template, ResultList):-
	dbfindall(DBId, Recipient, Template, ResultList).

itf_call(local, Pred):- 
	!, call(Pred).
itf_call(SiteId, Pred):-
	dbcall(SiteId, Pred).

ini_page_text_option(maintenance_title).
ini_page_text_option(maintenance_heading).
ini_page_text_option(webdb_title).
ini_page_text_option(webdb_heading).
ini_page_text_option(webdb_system_link).
ini_page_text_option(maintenance_system_link).
ini_page_text_option(stop_system_link).
ini_page_text_option(start_system_link).
ini_page_text_option(system_status_service).
ini_page_text_option(query_modify_service).
ini_page_text_option(modify_options_service).
ini_page_text_option(reload_templates_service).
ini_page_text_option(complex_query_service).
ini_page_text_option(create_table_service).
ini_page_text_option(select_table_to_edit_service).
ini_page_text_option(db_connect_service).

:- pred create_ini_page # "Creates both the initial page for database 
	maintenance and the front page for common user interaction.".

create_ini_page:-
% Maintenance part
       text_option(maintenance_title,_, Title),
       text_option(maintenance_heading,_, Heading),
       html_option(common_footer, [], Footer),
       html_option(main_page_heading, [Heading], Head),
       html_option(page_body, [[Head, Content, Footer]], Body),
       html_option(admin_content, [Links], Content),
       db_os(OS),
       (OS = 'Unix' ->
	 StartDbserver = 'admin/startdbserver.cgi',
	 Db_maintenance = 'admin/db_maintenance.cgi',
	 DbServerstop = 'admin/dbserverstop.cgi',
	 WebDBFront = 'webdb_front.cgi'
	;
	 StartDbserver = 'admin/startdbserver.cpx',
	 Db_maintenance = 'admin/db_maintenance.cpx',
	 DbServerstop = 'admin/dbserverstop.cpx',
	 WebDBFront = 'webdb_front.cpx'
	),
	maintenance_link(start_system_link, StartDbserver, StartLink),
	maintenance_link(maintenance_system_link, Db_maintenance, MaintenanceLink),	
	maintenance_link(stop_system_link, DbServerstop, StopLink),
       Links = [
		   StartLink,
		   MaintenanceLink,
		   StopLink
	        ],
       full_reply_page(Title, Body, [_|Page]),
       current_executable(Exe),
       absolute_file_name(Exe,'','','.',_,_,Dir),
       atom_concat(Dir,'/maintenance.html',Maintenance_file),
       umask(_, 8'113),
       open_output(Maintenance_file, IO),
       output_html(Page),
       close_output(IO),

% Client part
       text_option(webdb_title,_, WTitle),
       text_option(webdb_heading,_, WHeading),
       html_option(main_page_heading, [WHeading], WHead),
       html_option(page_body, [[WHead, WContent, Footer]], WBody),
       html_option(admin_content, [WebDBLink], WContent),
       maintenance_link(webdb_system_link, WebDBFront, WebDBLink),	
       full_reply_page(WTitle, WBody, [_|WPage]),
       db_dir(WDir),
       atom_concat(WDir,'/index.html',WebDB_file),
       umask(_, 8'113),
       open_output(WebDB_file, WIO),
       output_html(WPage),
       close_output(WIO).

maintenance_link(Name, CGI, Link) :-
        text_option(Name,_, Text),
        html_option(maintenance_link,[CGI,Text], Link).


platform_cgi(Service, CGI) :-
	db_os(OS),
	(OS = 'Unix' ->
	 atom_concat(Service, '.cgi', CGI)
	;
	 atom_concat(Service, '.cpx', CGI)).

% -----------------------------------------------------------------------------
 
:- comment(version(0*5+16,2003/06/17,16:26*30+'CEST'), "Deployment process
   improved and, in general, feedback resulting from AMOS applied (Jose
   Manuel Gomez Perez)").

:- comment(version_maintenance,dir('../version')).  :-
 comment(version(0*5+15,2003/01/08,13:32*17+'CET'), "Distribution updated
 for Windows (Jose Manuel Gomez Perez)").

:- comment(version(0*5+14,2002/04/15,11:27*06+'CEST'), "Corrected bug at
   numbers representation (Jose Manuel Gomez Perez)").

:- comment(version(0*5+13,2001/10/15,17:37*55+'CEST'), "Installation
   process and documentation updated (Jose Manuel Gomez Perez)").

:- comment(version(0*5+12,2001/03/23,17:48*27+'CET'), "Distribution process
   revised (Jose Manuel Gomez Perez)").

:- comment(version(0*5+0,2001/02/14,15:44*04+'CET'), "Due to changes in
   persdb, static persistent predicates also need to be declared
   dynamically.  (Jose Manuel Gomez Perez)").

:- comment(version(0*5+11,1999/09/16,12:52*05+'MEST'), "Added lock file to
   prevent the system from being started several times (Jose Manuel Gomez
   Perez)").

:- comment(version(0*5+10,1999/09/03,10:58*16+'MEST'), "Generic client
   interface added (Jose Manuel Gomez Perez)").

:- comment(version(0*5+8,1999/05/31,17:11*11+'MEST'), "Readmes now
automatically generated from documentation.  (Manuel Hermenegildo)").

:- comment(version(0*5+7,1999/04/23,13:14*44+'MEST'), "Naive graphical
statistics added (Jose Manuel Gomez Perez)").

:- comment(version(0*5+6,1999/04/21,11:58*39+'MEST'), "Text statistics
added (Jose Manuel Gomez Perez)").

:- comment(version(0*5+5,1999/04/20,13:09*27+'MEST'), "Infrastructure for
system status info added (Jose Manuel Gomez Perez)").

:- comment(version(0*5+4,1999/03/29,19:57*47+'MEST'), "HTML templates
revised (Jose Manuel Gomez Perez)").

:- comment(version(0*5+3,1999/03/23,18:58*21+'MET'), "Automatic link to
remote SQL databases through ODBC added (Jose Manuel Gomez Perez)").

:- comment(version(0*5+2,1999/02/10,12:29*33+'MET'), "Corrected lack of
generality at the edit tables interface (Jose Manuel Gomez Perez)").

:- comment(version(0*5+1,1999/02/10,12:28*51+'MET'), "Complex query with
restricted preds. completed (Jose Manuel Gomez Perez)").

:- comment(version(0*5+0,1998/07/23,13:26*00+'MET DST'), "Started
   autodocumentation.  (Manuel Hermenegildo)").

:- comment(version(0*1+1,1998/3/18), "Updated maintenance interface.  (Jose
   Manuel Gomez Perez)").

:- comment(version(0*1+0,1998/3/18), "Added recursive treatment of
   relations (base case).  (Jose Manuel Gomez Perez)").

:- comment(version(0*1+0,1998/2/20), "Started automatic
   documentation. (Jose Manuel Gomez Perez)").

:- comment(version(0*0+0,1998/2/20), "Started automatic
   documentation. (Jose Manuel Gomez Perez)").

% ----------------------------------------------------------------------------
