:- module(html_generation, 
	[
	    compose_page/4,
	    message_page/4,
	    form/3,
	    form_/3,
	    pred_to_form/4,
	    preds_to_form/4,
	    preds_to_data_table/3,
	    desc_to_heading/2,
	    name_list_to_table_heading/2,
	    data_to_table_row/2,
	    data_to_row_elements/2,
	    functor_to_field_names/3,
	    compose_table_row_field_names/4,
	    type_val_info/4,
	    type_info/3,
	    form_empty_list_values/1,
	    filter_empty/2,
	    get_one_pred_form_vals/5,
	    get_n_preds_form_vals/5
	], [assertions, basicmodes, pillow]).

:- use_module(library(terms), [atom_concat/2]).

:- use_module(html_generation_types).
:- use_module(html_templates_handle).
:- use_module(html_format_handle).


:- pred pred_to_form(+Pred, +HTMLDescription, +LFields, -Form) : term
	* pred_html_spec * list(atm) * html_term # "Takes a predicate
	@var{Pred} and the specification of its HTML representation
	@var{HTMLDescription} and generates a form @var{Form} with
	fields @var{Fields}.".

pred_to_form(Pred, HTMLDescription, LFields, Form):-
	html_option(general_table,['2',OutputFields], Form),
	HTMLDescription=..[_|Types],
	Pred=..[_|LValues],
	pred_to_form_field(LFields, Types, LValues, OutputFields).

pred_to_form_field([], [], [], []).
pred_to_form_field([F|Fs], [T|Ts], [V|Vs], [R|Rs]) :-
        type_val_info(T, V, Format, Description),
        html_format(Format, F, public, OutputField),
        html_option(data_row,[Description, OutputField], R),
        pred_to_form_field(Fs, Ts, Vs, Rs).


:- pred message_page(+Title, +Message, +BaseURL, -Response) : atm *
	atm * atm * html_term # "Uses @pred{compose_page/4} to
	generate a short message page".

message_page(Title, Message, BaseURL, Response) :-
	html_option(page_body, [[Head, Content]], Body),
	html_option(admin_heading, [Title], Head),
	html_option(admin_short_message,[Message], Mess),
	html_option(admin_content, [Mess], Content),
	compose_page(Title, Body, BaseURL, Response).


:- pred compose_page(+Title, +Body, +BaseURL, -Page) : atm * html_term
	* atm * html_term # "Generates an output page @var{Page}
	receiving as an input the page title @var{Title}, the main
	body @var{Body}, and a base url @var{BaseURL}".

compose_page(Title, Body, BaseURL, [cgi_reply,
                                    start,
				    base$[href=BaseURL],
				    title(Title),
				    Body,
				    end]).

:- pred form(+LFields, +HTMLDescription, -Form) : list(atm) *
	pred_html_spec * html_term # "Generates an empty form
	@var{Form} with fields @var{Fields} corresponding to a
	predicate whose HTML representation is specified in
	@var{HTMLDescription}. This form allows querying the above
	mentioned predicate. ".

form(LFields, HTMLDescription, Form) :-
        html_option(general_table,['2',InputFields], Form),
        HTMLDescription=..[_|Types],
        form_(LFields, Types, InputFields).

:- pred form_(+LFields, +HTMLDescriptionList, -Form) : list(atm) *
	pred_html_spec_list * html_term # "Generates an empty form
	@var{Form} with fields @var{Fields} corresponding to a
	predicate whose HTML representation of each of its arguments
	is specified in the list @var{HTMLDescriptionList}. This form
	allows querying the above mentioned predicate. ".

form_([], [], []).
form_([F|Fs], [T|Ts], [R|Rs]) :-
       type_info(T, Format, Description),
       html_format(Format, F, public, InputField),
       html_option(data_row, [Description, InputField], R),
       form_(Fs, Ts, Rs).


:- pred type_val_info(+Type, +Val, -Format, -Desc) : html_spec *
	field_basic_val * html_term * atm # "Given the HTML type
	@var{Type} of an argument of a predicate and the argument
	itself @var{Val}, obtains its corresponding HTML format
	@var{Format} and description @var{Desc}. ".

type_val_info(Type, Val, Format, Desc) :-
	Type =.. [T,Desc,_|Args],
	Format =.. [T,Val|Args].


:- pred type_info(+Type, -Format, -Desc) : html_spec * html_term * atm #
	"Given the HTML type @var{Type} of an argument of a predicate
	obtains its right HTML format @var{Format} and description
	@var{Desc}. ".

type_info(Type, Format, Desc) :-
        Type =.. [T,Desc|Args],
        Format =.. [T|Args].


name_arity(LPreds, Table, Arity) :-
	LPreds = [Pred|_],
	functor(Pred, Table, Arity).

:- pred preds_to_data_table(+LPreds, +LFieldsDesc, -Page) : list(term)
	* pred_html_spec_list * html_term # "Generates an HTML table
	where each row corresponds to one pred contained in
	@var{LPreds} formatted according to the types contained in
	@var{LFieldsDesc}".

preds_to_data_table([], [], []).
preds_to_data_table(LPreds, LFieldsDesc, Page):-
	name_arity(LPreds, _Table, Arity),
	preds_to_data_table_(Arity, LPreds, LFieldsDesc, Page).

preds_to_data_table_(Arity, LPreds, LFieldsDesc, Page):-
	desc_to_heading(LFieldsDesc, HeadData),
	data_to_table_row(HeadData, FirstRow),
	get_rows(LPreds, LFieldsDesc, RRows),
	Rows=[FirstRow|RRows],
	number_codes(Arity, A),
	atom_codes(AtomArity, A),
	html_option(general_table, [AtomArity, Rows], Page).

get_rows([], _, []). 
get_rows([P|Is], FieldsTypes,  [R1|Rs]):- 
	(P = pred(_N,I1) ->
	 I = I1
	;
	 I = P),
	I =.. [_|Fields],
	get_row(Fields, FieldsTypes, RowData),
	data_to_table_row(RowData, R1),
	get_rows(Is, FieldsTypes, Rs).

get_row([], _, []).
get_row([Field|Fs],[Type|Ts],[FinalField|Ds]):-
	(var(Field) ->
	 NewField = ''
	;
	 functor(Type, TypeName, _),
	 (TypeName = number ->
	  number_codes(Field, F),
	  atom_codes(NewField, F)
	 ;
	  NewField=Field)),
	html_option(table_element, [NewField], FinalField),
	get_row(Fs, Ts, Ds).


:- pred preds_to_form(+DisplayMode, +LPreds, +HTMLDescriptionList,
	-Table) : display_mode * list(term) * pred_html_spec_list *
	html_term # "Generates an HTML table @var{Table}where each row
	consists of form fields corresponding to one predicate
	contained in @var{LPreds}, formatted according to the HTML
	representation contained in
	@var{HTMLDescriptionList}. @var{DisplayMode} states whether
	the form is editable or not.".

preds_to_form(_,[],_,[]).
preds_to_form(DisplayMode, Preds, HTMLDescriptionList, Page):-
	desc_to_heading(HTMLDescriptionList, HeadData),
	data_to_table_row(HeadData, FirstRow),
	name_arity(Preds, Table, Arity),
	preds_to_form_(DisplayMode, Preds, Table, Arity, HTMLDescriptionList, 1, RRows),
	Rows=[FirstRow|RRows],
	number_codes(Arity, A),
	atom_codes(AtomArity, A),
	html_option(general_table, [AtomArity, Rows], Page).

preds_to_form_(_,[],_,_,_,_ ,[]).
preds_to_form_(DisplayMode,[Pred|Ps],Table,Arity,HTMLDescriptionList,RowIndex,[Row|Rs]):-
	Pred=..[_|Vals],
	get_row_field(DisplayMode, Vals, Table, RowIndex, Arity, HTMLDescriptionList, RowData),
	data_to_table_row(RowData, Row),
	NewRowIndex is RowIndex+1,
	preds_to_form_(DisplayMode,Ps,Table,Arity,HTMLDescriptionList,NewRowIndex,Rs).
	
get_row_field(DisplayMode,LVals,Table,RowIndex,Arity,HTMLDescriptionList,LRowFields):-
	compose_table_row_field_names(Table, RowIndex, Arity, LFields),
	output_field_table_elem(LFields, HTMLDescriptionList, LVals, DisplayMode, LRowFields).


:- pred get_n_preds_form_vals(+Input, +TI, +A, +NumberOfRows, -FieldValsList) : 
	form_dict * atm * int * int *list(list(field_basic_val)) 
        # "Obtains the values of a multirow form, given a Pillow Input
	@var{Input}, a table @var{TI}, its arity @var{A}, and the
	number of rows @var{NumberOfRows}.  The results are returned in a list 
        of lists, each corresponding to a form row. Normally used in 
        combination with @pred{preds_to_form/4}".

get_n_preds_form_vals(Input, Table, Arity, Length, LValues) :-
	get_n_preds_form_vals_(Input, Table, 1, Length, Arity, LValues).

get_n_preds_form_vals_(_, _, RowIndex, Length, _, []) :- 
	RowIndex > Length, !.
get_n_preds_form_vals_(Input, Table, RowIndex, Length, Arity, [LVals|LLVs]):- 
	get_vals(Input, Table, RowIndex, 1, Arity, LVals),
	NewRowIndex is RowIndex + 1,
	get_n_preds_form_vals_(Input, Table, NewRowIndex, Length, Arity, LLVs). 
get_vals(_, _, _, N, Arity, []):- N > Arity, !.
get_vals(Input, Table, RowIndex, N, Arity, [Val|Vs]):-
	number_codes(RowIndex, Rc),
	atom_codes(Ra, "_"||Rc),
	number_codes(N, Nc),
	atom_codes(Na, "_"||Nc),
	atom_concat(Ra, Na, Index),
	atom_concat(Table, Index, Field),
	get_form_value(Input, Field, Val0),
	filter_empty(Val0, Val),
	N1 is N + 1,
	get_vals(Input, Table, RowIndex, N1, Arity, Vs).

	
:- pred desc_to_heading(+HTMLDescriptions, -TableHeading) :
	pred_html_spec_list * html_term # "Takes a list of types
	specifying HTML representations @var{HTMLDescriptions} and
	generates an HTML table heading @var{TableHeading}".

desc_to_heading([], []).
desc_to_heading([HTMLDescription|HDs], [NewField|Fs]):-
	arg(1, HTMLDescription, Field),
	html_option(table_heading_elem, [Field], NewField),
	desc_to_heading(HDs, Fs).

:- pred name_list_to_table_heading(+Names, -TableHeading) : list(atm)
	* html_term # "Takes a list of atoms @var{Names} and generates
	an HTML table heading @var{TableHeading}.".

name_list_to_table_heading([], []).
name_list_to_table_heading([VarName|Vs], [NewVarName|Ns]):-
	html_option(table_heading_elem, [VarName], NewVarName),
	name_list_to_table_heading(Vs, Ns).

:- pred data_to_row_elements(+Elements, -RowElements) :
	list(field_basic_val) * html_term # "Takes a list of elements
	(atoms or numbers) @var{Elements}, transforms them into atoms
	and generates a list of table row elements @var{RowElements}
	with it.".

data_to_row_elements([], []).
data_to_row_elements([Arg|As],[FinalArg|FAs]):-
	atom(Arg),!,
	html_option(table_element, [Arg], FinalArg),
	data_to_row_elements(As, FAs).
data_to_row_elements([Number|As],[FinalArg|FAs]):-
	ground(Number), !,
	number_codes(Number, CN),
	atom_codes(Arg, CN),
	html_option(table_element, [Arg], FinalArg),
	data_to_row_elements(As, FAs).
data_to_row_elements([_Var|Vs],[FinalArg|FAs]):-
%	text_option(free_var_warning,_,Warning),
	html_option(table_element, [''], FinalArg),
	data_to_row_elements(Vs, FAs).


:- pred data_to_table_row(+RowData, -Row) : html_term * html_term #
	"Takes a list of table elements @var{RowData} and generates a
	table row with it @var{Row}".

data_to_table_row(RowData, Row):-
	html_option(table_row, [RowData], Row).

:- pred compose_table_row_field_names(+Table, +RowIndex, +Arity, -Fields) : 
	atm * int * int * list(atm) # "Generates a list of
	form field names @var{Fields} from the table name @var{Table},
	a number indicating the ordinal of the row @var{RowIndex} and
	the arity @var{Arity}, i.e. the number of fields, of the
	predicate represented.".

compose_table_row_field_names(Table, RowIndex, Arity, LFields) :-
	compose_table_row_field_names_(Table, RowIndex, 1, Arity, LFields).

compose_table_row_field_names_(_Table, _RowIndex, N, Arity, []):- N > Arity, !.
compose_table_row_field_names_(Table,  RowIndex, N, Arity, [Field|Fs]):-
	number_codes(RowIndex, Rc),
	atom_codes(Ra, "_"||Rc),
	number_codes(N, Nc),
	atom_codes(Na, "_"||Nc),
	atom_concat(Ra, Na, Index),
	atom_concat(Table, Index, Field),
	N1 is N + 1,
	compose_table_row_field_names_(Table,  RowIndex, N1, Arity, Fs).

output_field_table_elem([],[],[],_,[]).
output_field_table_elem([Field|Fs],[Type|Ts], [Val|Vs], DisplayMode, [RowField|RFs]):-
	type_val_info(Type, Val, Format, _),
	(Format=..[number, Value, N] -> 
	 (number(Val) ->
	  number_codes(Value, ValueList),
	  atom_codes(AValue, ValueList),
	  NewFormat=number(AValue, N)
	 ;
	  NewFormat=number('', N)
	 ),
	 html_format(NewFormat, Field, DisplayMode, RowFieldAux) 
	;
	html_format(Format, Field, DisplayMode, RowFieldAux)),
	html_option(table_element, [RowFieldAux], RowField),
	output_field_table_elem(Fs, Ts, Vs, DisplayMode, RFs).

:- pred form_empty_list_values(+ListValues) : list(field_basic_val) #
	"Test if the list of values @var{ListValues} of a form is
	empty.".

form_empty_list_values([]).
form_empty_list_values([FieldValue|LFieldValues]):-
        form_empty_value(FieldValue),
        form_empty_list_values(LFieldValues).


:- pred get_one_pred_form_vals(+Input, +TI, +A, -FieldList, -FieldValsList) :
	form_dict * atm * int * list(atm) * list(field_basic_val)
	# "Given a table @var{TI}, its arity @var{A} and a Pillow Input @var{Input} 
	obtains from a form the	fields and their values associated. ".

get_one_pred_form_vals(Input, TI, Arity, FieldList, FieldValsList) :-
	get_one_pred_form_vals_(1, Arity, TI, Input, FieldList, FieldValsList).

get_one_pred_form_vals_(N, Arity, _, _, [], []) :- N > Arity, !.
get_one_pred_form_vals_(N, Arity, TI, Input, [F|Fs], [FV|FVs]):-
	number_codes(N, S),
	atom_codes(U, "_"||S),
	atom_concat(TI, U, F),
	get_form_value(Input, F, FV),
	N1 is N+1,
	get_one_pred_form_vals_(N1, Arity, TI, Input, Fs, FVs).	


:- pred functor_to_field_names(+TI, +Arity, -FieldList) : atm * int *
	list(atm) # "Given a table @var{TI} and its arity @var{A},
	obtains the field names for the form associated. ".

functor_to_field_names(T, Arity, Fs) :-
	functor_to_field_names_(1, T, Arity, Fs).

functor_to_field_names_(N, _, Arity, []):- N > Arity, !.
functor_to_field_names_(N, T, Arity, [F|Fs]):-
	number_codes(N, S),
	atom_codes(U, "_"||S),
	atom_concat(T, U, F),
	N1 is N + 1,
	functor_to_field_names_(N1, T, Arity, Fs).

:- pred filter_empty(Value0, Value) : atm * atm # "Given a value
	@var{Value0} collected from a form outputs in @var{Value} a
	blank atom in case it is empty or the value itself in any
	other case".

filter_empty(Val0, '') :-
	form_empty_value(Val0), !.
filter_empty(Val, Val) .
