:- module(html_templates_handle, 
	[
	    html_option/3,
	    read_option_html/2
	], [assertions, basicmodes]).

:- use_module(library(file_utils), [file_to_string/2]).
:- use_module(library('pillow/html'), [html_template/3]).
:- use_module(library(terms), [atom_concat/2]).

:- data html_option__/4.             % Name, Explanation, Params, HTML_code
                                     % Correlates to the system html templates

:- pred html_option(+Name, +Vars, -HTMLOption) # "Fetches an html
	template @var{HTMLOption} called @var{Name} filling its vars
	with the values specified in @var{Vars}.".

html_option(Name, Vars, HTMLOption) :-
	html_option__(Name, _Title, Vars, HTMLOption).

:- pred read_option_html(+Name, +Dir) # "Transforms the html templates
     matching @var{Name} files stored in @var{Dir} into prolog terms
     which can be used by Pillow. ".

read_option_html(Name, Dir) :-
        option(Name, Params),
        create_dict(Params, Dict, Vars),
        atom_concat([Dir,Name, '.html'], HTML_file),
        ( file_to_string(HTML_file, Contents),
          html_template(Contents, HTML_terms, Dict),
          get_option(HTML_terms, Title, HTMLOption) -> true
        ; inform_user(['Bad/inexistent HTML file ',HTML_file]),
          fail
        ),
        retractall_fact(html_option__(Name, _, _, _)),
        assertz_fact(html_option__(Name, Title, Vars, HTMLOption)),
        fail.
read_option_html(_,_).

create_dict([], [], []).
create_dict([N|Ns], [N=V|NVs], [V|Vs]) :-
        create_dict(Ns, NVs, Vs).


get_option([E|Es], Title, HTMLOption) :- !,
        ( E = env(html,_,Contents) ->
              get_option2(Contents, Title, HTMLOption)
        ; get_option(Es, Title, HTMLOption)
        ).
get_option(env(html,_,Contents), Title, HTMLOption) :- !,
        get_option2(Contents, Title, HTMLOption).

get_option2([E|Es], Title, HTMLOption) :-
        ( E = env(head,_,InHead) ->
              get_title(InHead, Title),
              get_option3(Es, HTMLOption)
        ; E = env(title,_,[Title]) ->
              get_option3(Es, HTMLOption)
        ; get_option2(Es, Title, HTMLOption)
        ).

get_option3(env(body,[],HTMLOption), HTMLOption) :- !. % Body with no atts
get_option3(Es, HTMLOption) :- get_body_content(Es, HTMLOption), !.
get_option3(HTMLOption, HTMLOption).

get_title([env(title,_,[Title])|_], Title) :- !.
get_title([_|Es], Title) :-
        get_title(Es, Title).

get_body_content([E|_], HTMLOption) :-
        E = env(body,Atts,Contents), !,
        ( Atts = [] -> HTMLOption = Contents ; HTMLOption = E ).
get_body_content([_|Es], HTMLOption) :-
        get_body_content(Es, HTMLOption).


%%%%%%%%%%%%%%%%%%%%%%%%%% html templates description %%%%%%%%%%%%%%%%%%%%%%

option(page_body, [content]).
option(maintenance_link, [url,name]).
option(maintenance_footer, []).
option(admin_heading, [title]).
option(main_page_heading, [title]).
option(admin_content, [content]).
option(admin_short_message, [message]).
option(admin_footer, []).
option(common_footer, []).
option(ok_button, []).
option(reset_button, []).
option(form_buttons, [ok_button, reset_button]).
option(consult_table_heading, [table]).
option(consult_table_message, [content]).
option(match_message, [current,total]).
option(data_row, [table_heading_elem, table_element]).
option(control_row, [act, mode]).
option(action_display, [action, text, checked]).
option(service_display, [action, text, checked]).
option(general_choose, [text]).
option(deletion_warning, []).
option(option_entry, [name,text,value]).
option(options_body, [content]).
option(general_table, [arity, contents]).
option(no_frame_table, [arity, contents]).
option(table_row, [tuple]).
option(table_heading_elem, [value]).
option(table_element, [value]).	
option(table_added_message, [newtable]).
option(create_table_message, []).
option(general_sel_mess, [text]).
option(search_mode_display, [searchmode, text, checked]).
option(general_heading, [text]).
option(complex_query_instructions, []).
option(field_desc, [fdesc]).
option(db_connection_message, []).
option(big_bold_message, [text]).
option(general_add, [text]).
option(stats, [pic]).
